"""Workflow subpackage for local ScaleMS execution.

Execute subprocesses and functions in terms of the built-in asyncio framework.
Supports deferred execution by default, but close integration with the built-in
asyncio allows sensible behavior of concurrency constructs.

Example:
    python3 -m scalems.local my_workflow.py

"""

import asyncio
import importlib
import logging
import os
import pathlib
import weakref

from . import operations
from .. import execution as _execution
from .. import workflow as _workflow
from ..store import FileStore
from ..context import scoped_chdir
from ..exceptions import InternalError
from ..exceptions import MissingImplementationError
from ..exceptions import ProtocolError
from ..identifiers import TypeIdentifier
from ..subprocess._subprocess import SubprocessTask
from ..invocation import make_parser as _make_parser

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))


parser = _make_parser(__package__)


def workflow_manager(loop: asyncio.AbstractEventLoop):
    """Standard basic workflow context for local execution.

    Uses the asyncio module to allow commands to be staged as asyncio coroutines.

    There is no implicit OS level multithreading or multiprocessing.
    """
    return _workflow.WorkflowManager(loop=loop, executor_factory=executor_factory)


class _ExecutionContext:
    """Represent the run time environment for a managed workflow item."""

    def __init__(self, manager: _workflow.WorkflowManager, identifier: bytes):
        if manager is None:
            raise ProtocolError("Could not acquire reference to WorkflowManager. Stale reference?")
        self.workflow_manager: _workflow.WorkflowManager = manager
        self.identifier: bytes = identifier
        try:
            self.workflow_manager.item(self.identifier)
        except Exception as e:
            raise InternalError("Unable to access managed item.") from e
        self._task_directory = pathlib.Path(os.path.join(os.getcwd(), self.identifier.hex()))  #
        # TODO: Find a canonical way to get
        # the workflow directory.

    @property
    def task_directory(self) -> pathlib.Path:
        return self._task_directory

    def done(self, set_value: bool = None) -> bool:
        done_token = os.path.join(self.task_directory, "done")
        if set_value is not None:
            if set_value:
                pathlib.Path(done_token).touch(exist_ok=True)
            else:
                pathlib.Path(done_token).unlink(missing_ok=True)
        return os.path.exists(done_token)


class WorkflowUpdater(_execution.AbstractWorkflowUpdater):
    def __init__(self, runtime: _execution.RuntimeManager):
        self.executor = runtime

    async def submit(self, *, item: _workflow.Task) -> asyncio.Task:
        # TODO: Ensemble handling
        item_shape = item.description().shape()
        if len(item_shape) != 1 or item_shape[0] != 1:
            raise MissingImplementationError("Executor cannot handle multidimensional tasks yet.")

        task: asyncio.Task = await submit(item=item)
        return task


async def submit(item: _workflow.Task) -> asyncio.Task:
    key = item.uid()
    # Note that we could insert resource management here. We could create
    # tasks until we run out of free resources, then switch modes to awaiting
    # tasks until resources become available, then switch back to placing tasks.
    execution_context = _ExecutionContext(item.manager(), key)
    if execution_context.done():
        if isinstance(key, bytes):
            id = key.hex()
        else:
            id = str(key)
        logger.info(f"Skipping task that is already done. ({id})")
        # TODO: Update local status and results.
    else:
        assert not item.done()
        assert not item.manager().item(key).done()
        task_type: TypeIdentifier = item.description().type()
        task = asyncio.create_task(_execute_item(task_type=task_type, item=item, execution_context=execution_context))
        return task


# TODO: return an execution status object?
async def _execute_item(
    task_type: TypeIdentifier, item: _workflow.Task, execution_context: _ExecutionContext  # noqa: C901
):
    # TODO: Automatically resolve resource types.
    if task_type == TypeIdentifier(("scalems", "subprocess", "SubprocessTask")):
        task_type = SubprocessTask()

        # TODO: Use abstract input factory.
        logger.debug("Resolving input for {}".format(str(item)))
        input_type = task_type.input_type()
        input_record = input_type(**item.input)
        input_resources = operations.input_resource_scope(context=execution_context, task_input=input_record)

        # We need to provide a scope in which we guarantee the availability of resources,
        # such as temporary files provided for input, or other internally-generated
        # asyncio entities.
        async with input_resources as subprocess_input:
            logger.debug("Creating coroutine for {}".format(task_type.__class__.__name__))
            # TODO: Use abstract task factory.
            coroutine = operations.subprocessCoroutine(context=execution_context, signature=subprocess_input)
            logger.debug("Creating asyncio Task for {}".format(repr(coroutine)))
            awaitable = asyncio.create_task(coroutine)

            # TODO: Use abstract results handler.
            logger.debug("Waiting for task to complete.")
            result = await awaitable
            subprocess_exception = awaitable.exception()
            if subprocess_exception is not None:
                logger.exception("subprocess task raised exception {}".format(str(subprocess_exception)))
                raise subprocess_exception
            logger.debug("Setting result for {}".format(str(item)))
            item.set_result(result)
            # TODO: Need rigorous task state machine.
            execution_context.done(item.done())
    else:
        # TODO: Process pool?
        # The only way we can be sure that tasks will not detect and use the working
        # directory of the main script's interpreter process is to actually
        # change directories. However, there isn't a good way to do that with concurrency.
        # We can either block concurrent execution within the changed directory,
        # or we can do the directory changing in subprocesses.
        #
        # Here we use a non-asynchronous context manager to avoid yielding
        # to the asyncio event loop. However, if there are other threads in the
        # current process that release the GIL, this is not safe.
        try:
            task_dir = execution_context.task_directory
            if not task_dir.exists():
                task_dir.mkdir()
            logger.debug(f"Entering {task_dir}")
            # Note: This limits us to one task-per-process.
            # TODO: Use one (or more) subprocesses per task, setting the working
            #  directory.
            # The forking may be specialized to the task type, or we can default
            # to just always forking the Python interpreter.
            with scoped_chdir(task_dir):
                # TODO: Use implementation registry.
                # We should have already checked for this...
                namespace_end = len(task_type.scoped_name())
                implementation = None
                while namespace_end >= 1:
                    try:
                        module = ".".join(task_type.scoped_name()[:namespace_end])
                        logger.debug(f"Looking for importable module: {module}")
                        implementation = importlib.import_module(module)
                    except ImportError:
                        implementation = None
                    else:
                        break
                    # Try the parent namespace?
                    namespace_end -= 1
                if namespace_end < 1:
                    raise InternalError("Bad task_type_identifier.")
                logger.debug(f'Looking for "scalems_helper in {module}.')
                helper = getattr(implementation, "scalems_helper", None)
                if not callable(helper):
                    raise MissingImplementationError(
                        "Executor does not have an implementation for {}".format(str(task_type))
                    )
                else:
                    logger.debug("Helper found. Passing item to helper for execution.")
                    # Note: we need more of a hook for two-way interaction here.
                    try:
                        local_resources = execution_context.workflow_manager
                        result = helper(item, context=local_resources)
                        item.set_result(result)
                    except Exception as e:
                        raise e
                    execution_context.done(item.done())
        except Exception as e:
            logger.exception(f"Badly handled exception: {e}")
            raise e


def executor_factory(manager: _workflow.WorkflowManager, params=None):
    if params is not None:
        raise TypeError("This factory does not accept a Configuration object.")
    executor = LocalExecutor(
        editor_factory=weakref.WeakMethod(manager.edit_item),
        datastore=manager.datastore(),
        loop=manager.loop(),
        dispatcher_lock=manager._dispatcher_lock,
    )
    return executor


class LocalExecutor(_execution.RuntimeManager):
    """Client side manager for work dispatched locally."""

    def __init__(
        self,
        *,
        editor_factory=None,
        datastore: FileStore = None,
        loop: asyncio.AbstractEventLoop,
        configuration=None,
        dispatcher_lock=None,
    ):
        """Create a client side execution manager.

        Initialization and deinitialization occurs through
        the Python (async) context manager protocol.
        """
        super().__init__(
            editor_factory=editor_factory,
            datastore=datastore,
            loop=loop,
            configuration=configuration,
            dispatcher_lock=dispatcher_lock,
        )

    def updater(self) -> WorkflowUpdater:
        return WorkflowUpdater(runtime=self)

    async def runtime_startup(self) -> asyncio.Task:
        runner_started = asyncio.Event()
        runner_task = asyncio.create_task(_execution.manage_execution(executor=self, processing_state=runner_started))
        await runner_started.wait()
        return runner_task



"""Support for scalems on radical.pilot.raptor.

Define the connective tissue for SCALE-MS tasks embedded in rp.Task arguments.

Provide the RP Raptor Master and Worker details. Implement the master task
script as a package entry point that we expect to be executable from the
command line in a virtual environment where the scalems package is installed,
without further modification to the PATH.

The client should be reasonably certain that the target environment has a
compatible installation of RP and scalems. A rp.raptor.Master task script is
installed with the scalems package. The script name is provide by the
module function :py:func:`~scalems.radical.raptor.master_script()`, and will
be resolvable on the PATH for a Python interpreter process in an
environment that has the `scalems` package installed.

We should try to keep this module as stable as possible so that the run time
interface provided by scalems to RP is robust, and the entry point scripts
change as little as possible across versions.
scalems.radical runtime details live in runtime.py.

As of :py:mod:`radical.pilot` version 1.18, TaskDescription supports
a *mode* field and several additional fields. The extended schema for the TaskDescription
depends on the value of *mode*. The structured data provided to the executor callable
is composed from these additional fields according to the particular mode.

We use the :py:data:`radical.pilot.task_description.TASK_FUNCTION` mode,
specifying a *function* field to name a callable in the (global) namespace
accessible by the worker process.

We populate the worker global namespace with imported callables in the module
file from which Raptor imports our :py:class:`radical.pilot.raptor.MPIWorker` subclass,
:py:class:`~scalems.radical.raptor.ScaleMSWorker`.

The callable for *call* accepts ``*args`` and ``**kwargs``, which are extracted
from fields in the TaskDescription. Since we derive from MPIWorker, an mpi4py
communicator is inserted at ``args[0]``.

Protocol
--------

As of RP 1.18, the :py:mod:`radical.pilot.raptor` interface is not documented.
The following diagrams illustrate the approximate architecture of a Raptor Session.

.. uml::

    title RP Raptor client Session

    box "RP client"
    participant TaskManager
    end box

    'queue "Taskmanager: Scheduler and Input Queue" as Queue
    queue "Scheduler/Input Queue" as Queue

    -> TaskManager: submit(master_task_description)
    activate TaskManager

    'note left
    note over TaskManager
     TaskDescription uses `mode=rp.RAPTOR_MASTER`.
     Task args provide details for the script
     so that a Master can start appropriate Workers.
    end note

    TaskManager -> Queue: task_description
    activate Queue
    deactivate TaskManager
    Queue <-
    Queue -->
    deactivate Queue

    -> TaskManager: submit(task_description)
    activate TaskManager

    'note left
    note over TaskManager
     TaskDescription names a Master uid in *scheduler* field.
    end note

    TaskManager -> Queue: task_description
    activate Queue
    deactivate TaskManager
    Queue <-
    Queue -->
    deactivate Queue
    ...
    TaskManager ->
    TaskManager <-- : master task results

A "master task" is an *executable* task (*mode* = ``rp.RAPTOR_MASTER``)
in which the script named by
:py:data:`radical.pilot.TaskDescription.executable`
manages the life cycle of a :py:class:`radical.pilot.raptor.Master`
(or subclass instance).
As of RP 1.14, the protocol is as follows.

.. uml::

    title raptor master task lifetime management

    participant "master task"

    -> "master task" : stage in

    create Master as master
    "master task" -> master: create Master(cfg)
    "master task" -> master: Master.submit_workers(descr=descr, count=n_workers)
    "master task" -> master: Master.start()
    alt optional hook for self-submitting additional tasks
    "master task" -> master: Master.submit_tasks(tasks)
    end
    queue scheduler
    scheduler -\\ master : request_cb
    scheduler -\\ master : result_cb
    master -> master: Master.stop() (optional)
    "master task" -> master: Master.join()

    [<-- "master task" : stage out


The *cfg* argument to :py:class:`radical.pilot.raptor.Master` does not currently
appear to be required.

*scalems* encodes worker requirements on the client side. The *scalems* master
script decodes the client-provided requirements, combines the information with
run time details and work load inspection, and produces the WorkerDescription
with which to :py:func:`~radical.pilot.raptor.Master.submit_workers()`.

.. uml::

    class RaptorWorkerConfig {
     count
     descr
    }

    class WorkerDescription

    RaptorWorkerConfig *--> "descr" WorkerDescription

    namespace scalems.radical {
     class MasterTaskConfiguration

     class ClientWorkerRequirements

     class ScaleMSMaster

     ScaleMSMaster --> MasterTaskConfiguration : launches with

     MasterTaskConfiguration *-- ClientWorkerRequirements

     ClientWorkerRequirements -right-> .RaptorWorkerConfig : worker_description()
    }

    namespace radical.pilot.raptor {
     class Master {
      submit_workers()
     }
     Master::submit_workers .> .RaptorWorkerConfig
    }

    scalems.radical.ScaleMSMaster -up-|> radical.pilot.raptor.Master


.. warning:: The data structure for *descr* may still be evolving.
    See https://github.com/radical-cybertools/radical.pilot/issues/2731

.. uml::

    title scalems raptor runtime management
    !pragma teoz true
    box "client"
    box "scalems.radical" #honeydew
    participant RPDispatchingExecutor as client_runtime
    end box

    box "radical.pilot" #White
    participant task_manager
    end box
    end box

    queue "RP runtime" as scheduler

    -> client_runtime : runtime_startup()
    activate client_runtime
    client_runtime -> client_runtime : _get_scheduler()
    activate client_runtime
    client_runtime -> client_runtime : master_script()
    client_runtime -> client_runtime : master_input()
    activate client_runtime
    client_runtime -> client_runtime : worker_requirements()
    activate client_runtime

    note over client_runtime
    TODO: Allocate Worker according to workload
    through a separate call to the running Master.
    end note

    return ClientWorkerRequirements
    return MasterTaskConfiguration

    client_runtime -> client_runtime: launch Master Task
    activate client_runtime

    note over client_runtime, task_manager
    We currently require a pre-existing venv for Master task.
    TODO(#90,#141): Allow separate venv for Worker task and Tasks.
    end note

    client_runtime -> task_manager : submit master task
    task_manager -\\ scheduler

    note over task_manager, scheduler
    Master cfg is staged with TaskDescription.
    end note

    deactivate client_runtime

    note over client_runtime, task_manager
    TODO(#92,#105): Make sure the Worker starts successfully!!!
    end note

    return master task
    return dispatching context


TODO
----
Pass input and output objects more efficiently.

Worker processes come and go,
and are not descended from Master processes (which may not even be on the same node),
so we can't generally pass objects by reference or even in shared memory segments.
However, we can use optimized ZMQ facilities for converting network messages
directly to/from memory buffers.

In other cases, we can use the scalems typing system to specialize certain types
for optimized (de)serialization from/to file-backed storage or streams.

See https://github.com/SCALE-MS/scale-ms/issues/106

TODO
----
Get a stronger specification of the RP Raptor interface.

See https://github.com/radical-cybertools/radical.pilot/issues/2731

"""
from __future__ import annotations

__all__ = (
    "ClientWorkerRequirements",
    "MasterTaskConfiguration",
    "master_script",
    "master_input",
    "master",
    "worker_requirements",
    "worker_description",
    "SoftwareCompatibilityError",
    "ScaleMSWorker",
    "ScaleMSMaster",
    "ScalemsRaptorWorkItem",
    "WorkerDescription",
    "RaptorWorkerConfig",
)

import abc
import argparse
import asyncio
import contextlib
import dataclasses
import functools
import importlib
import importlib.metadata
import json
import os
import sys
import tempfile
import typing
import warnings
import weakref
import zlib
from importlib.machinery import ModuleSpec
from importlib.util import find_spec
from collections.abc import Generator

import packaging.version

if typing.TYPE_CHECKING:
    try:
        from mpi4py.MPI import Comm
    except ImportError:
        Comm = None


# We import rp before `logging` to avoid warnings when rp monkey-patches the
# logging module. This `try` suite helps prevent auto-code-formatters from
# rearranging the import order of built-in versus third-party modules.
try:
    import radical.pilot as rp
except (ImportError,):
    warnings.warn("RADICAL Pilot installation not found.")

import scalems.exceptions
import scalems.radical
import scalems.messages
import scalems.file as _file
import scalems.store as _store

# QUESTION: How should we approach logging? To what degree can/should we integrate
# with rp logging?
import logging

logger = logging.getLogger(__name__)
# Note that we have not restricted propagation or attached a handler, so the messages
# to `logger` will end up going to the stderr of the process that imported this module
# (e.g. the master task or the Worker task) and appearing in the corresponding
# Task.stderr. This is available on the client side in the case of the master task,
# but only to the Master itself for the worker task.
# We should probably either add a specific LogHandler or integrate with the
# RP logging and Component based log files.
# See also
# * https://github.com/SCALE-MS/scale-ms/discussions/261
# * https://github.com/SCALE-MS/scale-ms/issues/255


@dataclasses.dataclass(frozen=True)
class BackendVersion:
    """Identifying information for the computing backend."""

    name: str
    """Identifying name (presumably a module name)."""

    version: str
    """Implementation revision identifier as a PEP 440 compatible version string."""


backend_version = BackendVersion(name="scalems.radical.raptor", version="0.0.0")

# TODO: Where does this identifier belong?
api_name = "scalems_v0"
"""Key for dispatching raptor Requests.

We can use this to identify the schema used for SCALE-MS tasks encoded in
arguments to raptor :py:data:`~radical.pilot.TASK_FUNCTION` mode executor.
"""

CPI_MESSAGE = "scalems.cpi"
"""Flag for scalems messages to be treated as CPI calls.

Used in the :py:attr:`~radical.pilot.TaskDescription.mode` field to indicate that the
object should be handled through the SCALEMS Compute Provider Interface machinery.
"""


class CpiCommand(abc.ABC):
    _registry: typing.MutableMapping[str, typing.Type["CpiCommand"]] = weakref.WeakValueDictionary()

    @classmethod
    @abc.abstractmethod
    def launch(cls, manager: ScaleMSMaster, task: TaskDictionary):
        """Process the RP Task as a CPI Command.

        Called in ScaleMSMaster.request_cb().
        """
        ...

    @classmethod
    @abc.abstractmethod
    def result_hook(cls, manager: ScaleMSMaster, task: TaskDictionary):
        """Called during Master.result_cb."""

    @typing.final
    @classmethod
    def get(cls, command: scalems.messages.Command):
        return CpiCommand._registry[command.__class__.__qualname__]

    @classmethod
    @abc.abstractmethod
    def command_class(cls) -> str:
        """The qualified name of the associated `scalems.messages` Command class."""
        ...

    def __init_subclass__(cls, **kwargs):
        if cls is not CpiCommand:
            CpiCommand._registry[cls.command_class()] = cls
        super().__init_subclass__(**kwargs)


class CpiStop(CpiCommand):
    """Provide implementation for StopCommand in RP Raptor."""

    @classmethod
    def command_class(cls) -> str:
        return scalems.messages.StopCommand.__qualname__

    @classmethod
    def launch(cls, manager: ScaleMSMaster, task: TaskDictionary):
        logger.debug("CPI STOP issued.")
        # TODO: Wait for pending tasks to complete and shut down Worker(s)
        task["stderr"] = ""
        task["stdout"] = ""
        task["exit_code"] = 0
        manager.cpi_finalize(task)

    @classmethod
    def result_hook(cls, manager: ScaleMSMaster, task: TaskDictionary):
        manager.stop()
        logger.debug(f"Finalized {str(task)}.")


class CpiHello(CpiCommand):
    """Provide implementation for HelloCommand in RP Raptor."""

    @classmethod
    def launch(cls, manager: ScaleMSMaster, task: TaskDictionary):
        logger.debug("CPI HELLO in progress.")
        task["stderr"] = ""
        task["exit_code"] = 0
        task["stdout"] = repr(backend_version)
        # TODO: scalems object encoding.
        task["return_value"] = dataclasses.asdict(backend_version)
        logger.debug("Finalizing...")
        manager.cpi_finalize(task)

    @classmethod
    def result_hook(cls, manager: ScaleMSMaster, task: TaskDictionary):
        logger.debug(f"Finalized {str(task)}.")

    @classmethod
    def command_class(cls) -> str:
        return scalems.messages.HelloCommand.__qualname__


class CpiAddItem(CpiCommand):
    """Add an item to the managed workflow record.

    Descriptions of Work with no dependencies will be immediately scheduled for
    execution (converted to RP tasks and submitted).

    TBD: Data objects, references to existing data, completed tasks.
    """

    def __init__(self, add_item: scalems.messages.AddItem):
        encoded_item = add_item.encoded_item
        item_dict = json.loads(encoded_item)
        self.work_item = ScalemsRaptorWorkItem(
            func=item_dict["func"],
            module=item_dict["module"],
            args=item_dict["args"],
            kwargs=item_dict["kwargs"],
            comm_arg_name=item_dict.get("comm_arg_name", None),
        )

    @classmethod
    def launch(cls, manager: ScaleMSMaster, task: TaskDictionary):
        """Repackage a AddItem command as a rp.TaskDescription for submission to the Worker.

        Note that we are not describing the exact function call directly,
        but an encoded function call to be handled by the `run_in_worker`
        dispatching function. The arguments will be serialized
        together with the function code object into the
        *work_item* key word argument for the RP Task.

        The Master.request_cb() submits individual scalems tasks with this function.
        More complex work is deserialized and managed by ScaleMSMaster using
        the `raptor_work_deserializer` function, first.

        See Also:
            `scalems.radical.raptor.ScaleMSWorker.run_in_worker()`
        """
        # submit the task and return its task ID right away,
        # then allow its result to either just be used to
        #   * trigger dependent work,
        #   * only appear in the Master report, or
        #   * be available to follow-up LRO status checks.
        add_item = typing.cast(
            scalems.messages.AddItem, scalems.messages.Command.decode(task["description"]["metadata"])
        )
        encoded_item = add_item.encoded_item
        # We do not yet use a strongly specified object schema. Just accept a dict.
        item_dict = json.loads(encoded_item)
        # Decouple the serialization schema since it is not strongly specified or robust.
        work_item = ScalemsRaptorWorkItem(
            func=item_dict["func"],
            module=item_dict["module"],
            args=item_dict["args"],
            kwargs=item_dict["kwargs"],
            comm_arg_name=item_dict.get("comm_arg_name", None),
        )
        # TODO(#277): Convert abstract inputs to concrete values. (More management interface.)
        # TODO(#277): Check dependencies and runnability before submitting.

        fingerprint = zlib.crc32(json.dumps(work_item, sort_keys=True).encode("utf8"))
        # TODO: More robust fingerprint. Ref scalems.serialization and scalems.store
        # TODO: Check for duplicates.
        scalems_task_id = f"scalems-task-{fingerprint}"

        scalems_task_description = rp.TaskDescription(
            _RaptorTaskDescription(
                uid=scalems_task_id,
                # Note that (as of this writing) *executable* is required but unused for Raptor tasks.
                executable="scalems",
                scheduler=manager.uid,
                # Note we could use metadata to encode additional info for ScaleMSMaster.request_cb,
                # but it is not useful for execution of the rp.TASK_FUNCTION.
                metadata=None,
                mode=rp.TASK_FUNCTION,
                function="run_in_worker",
                args=[],
                kwargs={"work_item": work_item},
            )
        )
        # Note: There is no `Task` object returned from `Master.submit_tasks()`
        # `Task` objects are currently only available on the client side;
        # the agent side handles task dicts---which the master will receive through the *request_cb()*.
        manager.submit_tasks(scalems_task_description)
        logger.debug(f"Submitted {str(scalems_task_description)} in support of {str(task)}.")

        # Record task metadata and track.
        task["return_value"] = scalems_task_id
        task["stdout"] = scalems_task_id
        task["exit_code"] = 0
        manager.cpi_finalize(task)

    @classmethod
    def result_hook(cls, manager: ScaleMSMaster, task: TaskDictionary):
        logger.debug(f"Finalized {str(task)}.")

    @classmethod
    def command_class(cls) -> str:
        return scalems.messages.AddItem.__qualname__


EncodableAsDict = typing.Mapping[str, "Encodable"]
EncodableAsList = typing.List["Encodable"]
Encodable = typing.Union[str, int, float, bool, None, EncodableAsDict, EncodableAsList]


# def object_decoder(obj: dict):
#     """Provide the object_hook callback for a JSONDecoder."""


@functools.singledispatch
def object_encoder(obj) -> Encodable:
    """Provide the *default* callback for JSONEncoder."""
    raise TypeError(f"No decoder for {obj.__class__.__qualname__}.")


# def get_decoder() -> json.JSONDecoder:
#     """Get a JSONDecoder instance extended for types in this module."""
#     decoder = json.JSONDecoder(object_hook=object_decoder)
#     return decoder
#
#
# def get_encoder() -> json.JSONEncoder:
#     """Get a JSONEncoder instance extended for types in this module."""
#     encoder = json.JSONEncoder(default=object_encoder)
#     return encoder


# @object_encoder.register
# def _(obj: rp.TaskDescription) -> dict:
#     return obj.as_dict()


@dataclasses.dataclass(frozen=True)
class ClientWorkerRequirements:
    """Client-side details to inform worker provisioning.

    This structure is part of the `scalems.radical.raptor.MasterTaskConfiguration`
    provided to the master script. The master script uses this information
    when calling `worker_description()`.

    Use the :py:func:`worker_requirements()` creation function for a stable interface.
    """

    cpu_processes: int
    """Number of ranks in the Worker MPI context.

    TODO: We need to account for cores-per-process.
    """

    gpus_per_process: int = 0
    """GPUs per Worker rank.

    TODO: Reconcile with upcoming support for fractional ratios.
    """

    pre_exec: tuple[str] = ()
    """Lines of shell expressions to be evaluated before launching the worker process."""

    named_env: str = None
    """A registered virtual environment known to the raptor manager.

    Warnings:
        Not tested. Support has been delayed indefinitely.
    """


class _MasterTaskConfigurationDict(typing.TypedDict):
    """A `MasterTaskConfiguration` encoded as a `dict`."""

    worker: dict  # dict-encoded ClientWorkerRequirements.
    versioned_modules: typing.List[typing.Tuple[str, str]]


@dataclasses.dataclass
class MasterTaskConfiguration:
    """Input to the script responsible for the RP raptor Master.

    .. todo:: Check the schema for RP.
        See https://github.com/radical-cybertools/radical.pilot/issues/2731

    Produced on the client side with `master_input`. Deserialized in the
    master task (`master`) to get a `RaptorWorkerConfig`.
    """

    worker: ClientWorkerRequirements
    """Client-side details to inform worker provisioning.

    Generated by `worker_requirements()`.
    Used to prepare input for `worker_description()`.
    """

    versioned_modules: typing.List[typing.Tuple[str, str]]
    """List of name, version specifier tuples for required modules."""

    @classmethod
    def from_dict(cls, obj: _MasterTaskConfigurationDict):
        """Decode from a native dict.

        Support deserialization, such as from a JSON file in the master task.
        """
        return cls(worker=ClientWorkerRequirements(**obj["worker"]), versioned_modules=list(obj["versioned_modules"]))


@object_encoder.register
def _(obj: MasterTaskConfiguration) -> dict:
    return dataclasses.asdict(obj)


@functools.cache
def master_script() -> str:
    """Get the name of the RP raptor master script.

    The script to run a RP Task based on a rp.raptor.Master is installed
    with :py:mod`scalems`. Installation configures an "entry point" script
    named ``scalems_rp_master``, but for generality this function should
    be used to get the entry point name.

    Before returning, this function confirms the availability of the entry point
    script in the current Python environment. A client should arrange for
    the script to be called in the execution environment and to confirm
    that the (potentially remote) entry point matches the expected API.

    Returns:
        str: Installed name of the entry point wrapper for :py:func:`~scalems.radical.raptor.master()`

    """
    try:
        # TODO(Pyhon 3.10): Use importlib.metadata.entry_points(group='console_scripts')?
        import pkg_resources
    except ImportError:
        pkg_resources = None
    _master_script = "scalems_rp_master"
    if pkg_resources is not None:
        # It is not hugely important if we cannot perform this test.
        # In reality, this should be performed at the execution site, and we can/should
        # remove the check here once we have effective API compatibility checking.
        # See https://github.com/SCALE-MS/scale-ms/issues/100
        assert pkg_resources.get_entry_info("scalems", "console_scripts", "scalems_rp_master").name == _master_script
    return _master_script


async def master_input(
    *, filestore: _store.FileStore, worker_pre_exec: list, worker_venv: str
) -> _file.AbstractFileReference:
    """Provide the input file for a SCALE-MS Raptor Master script.

    Args:
        worker_venv (str) : Directory path for the Python virtual environment in which
            to execute the worker.
        worker_pre_exec (list[str]): List of shell command lines to execute in the worker
            environment before its launch script.
        filestore: (local) FileStore that will manage the generated AbstractFileReference.

    """
    if not isinstance(filestore, _store.FileStore) or filestore.closed or not filestore.directory.exists():
        raise ValueError(f"{filestore} is not a usable FileStore.")

    # This is the initial Worker submission. The Master may submit other workers later,
    # but we should try to make this one as usable as possible.
    task_metadata = worker_requirements(pre_exec=worker_pre_exec, worker_venv=worker_venv)

    # TODO(#248): Decide on how to identify the workers from the client side.
    # We can't know the worker identity until the master task has called submit_workers()
    # filestore.add_task(worker_identity, **task_metadata)

    # TODO(#141): Add additional dependencies that we can infer from the workflow.
    versioned_modules = (("mpi4py", "3.0.0"), ("scalems", scalems.__version__), ("radical.pilot", rp.version))

    configuration = MasterTaskConfiguration(worker=task_metadata, versioned_modules=list(versioned_modules))

    # Make sure the temporary directory is on the same filesystem as the local workflow.
    tmp_base = filestore.directory
    # TODO(Python 3.10) Use `ignore_cleanup_errors=True`
    tmpdir_manager = tempfile.TemporaryDirectory(dir=tmp_base)
    with tmpdir_manager as tmpdir:
        # Serialize the configuration to a temporary file, then add it to the
        # FileStore to get a fingerprinted, tracked file.
        config_file_name = "raptor_scheduler_config.json"
        config_file_path = os.path.join(tmpdir, config_file_name)
        with open(config_file_path, "w") as fh:
            json.dump(configuration, fh, default=object_encoder, indent=2)
        file_description = _file.describe_file(config_file_path, mode="r")
        add_file_task = asyncio.create_task(
            _store.get_file_reference(file_description, filestore=filestore), name="get-config-file-reference"
        )
        await asyncio.wait((add_file_task,), return_when=asyncio.FIRST_EXCEPTION)
        assert add_file_task.done()
        try:
            tmpdir_manager.cleanup()
        except OSError:
            logger.exception(f"Errors occurred while cleaning up {tmpdir}.")
    if isinstance(add_file_task.exception(), scalems.exceptions.DuplicateKeyError):
        checksum = await asyncio.to_thread(file_description.fingerprint)
        key = checksum.hex()
        # Note: the DuplicateKeyError could theoretically be the result of a
        # pre-existing file with the same internal path but a different key, but
        # such a condition would probably represent an internal error (bug).
        # TODO: Migrate away from container behavior of FileStore.
        return filestore[key]
    else:
        return add_file_task.result()


def worker_requirements(*, pre_exec: typing.Iterable[str], worker_venv: str) -> ClientWorkerRequirements:
    """Get the requirements for the work load, as known to the client.

    TODO: Inspect workflow to optimize reusability of the initial Worker submission.
    """
    # TODO: calculate cores.
    # rp_config = scalems.radical.configuration()
    # pilot_description = rp_config.rp_resource_params["PilotDescription"]
    # available_cores = pilot_description.get("cores")
    # if not available_cores:
    #     cores_per_node =
    #     ...
    cores_per_worker = 1
    gpus_per_worker = 0

    workload_metadata = ClientWorkerRequirements(
        named_env=worker_venv,
        pre_exec=tuple(pre_exec),
        cpu_processes=cores_per_worker,
        gpus_per_process=gpus_per_worker,
    )

    return workload_metadata


parser = argparse.ArgumentParser(description="Command line entry point for RP Raptor master task.")
parser.add_argument("file", type=str, help="Input file (JSON) for configuring ScaleMSMaster instance.")


def master():
    """Entry point for raptor.Master task.

    This function implements the :command:`scalems_rp_master` entry point script
    called by the RADICAL Pilot executor to provide the raptor master task.

    During installation, the `scalems` build system creates a :file:`scalems_rp_master`
    `entry point <https://setuptools.pypa.io/en/latest/pkg_resources.html#entry-points>`__
    script that provides command line access to this function. The Python signature
    takes no arguments because the function processes command line arguments from `sys.argv`

    .. uml::

        title scalems raptor master task lifetime
        !pragma teoz true

        queue "RP runtime" as scheduler

        box execution
        box "scalems.radical" #honeydew
        participant "master task"
        participant ScaleMSMaster as master
        end box
        participant worker.py
        end box

        scheduler -> "master task" : stage in
        ?-> "master task" : scalems_rp_master
        activate "master task"

        note over "master task" : "TODO: Get asyncio event loop?"
        note over "master task" : "TODO: Initialize scalems FileStore"

        "master task" -> "master task" : from_dict
        activate "master task"
        return MasterTaskConfiguration
        "master task" -> master **: create ScaleMSMaster()

        "master task" -> "master task" : with configure_worker()
        activate "master task"
        "master task" -> master : configure_worker()
        activate master
        master -> worker.py ** : with _worker_file()
        activate master
        master -> master : _configure_worker()
        activate master
        master -> master : worker_description()
        activate master
        return
        return WorkerDescription
        deactivate master
        return RaptorWorkerConfig
        'master --> "master task" : RaptorWorkerConfig
        'deactivate master

        "master task" -> master: submit_workers(**config)
        "master task" -> master: start()
        note over "master task" : "TODO: wait for worker"

        ...
        scheduler -\\ master : request_cb
        scheduler -\\ master : result_cb
        ...

        "master task" -> master : ~__exit__
        activate master
        master -> master : ~__exit__
        activate master
        master --> master : stop worker (TODO)
        master -> worker.py !! : delete
        deactivate master
        master -> master: stop() (TBD)
        "master task" -> master: join()
        deactivate master
        deactivate "master task"

        scheduler <-- "master task" : stage out
        deactivate "master task"

    """
    # TODO: Configurable log level?
    logger.setLevel(logging.DEBUG)
    character_stream = logging.StreamHandler()
    character_stream.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    character_stream.setFormatter(formatter)
    logger.addHandler(character_stream)

    file_logger = logging.FileHandler("scalems.radical.raptor.log")
    file_logger.setLevel(logging.DEBUG)
    file_logger.setFormatter(formatter)
    logging.getLogger("scalems").addHandler(file_logger)

    if not os.environ["RP_TASK_ID"]:
        warnings.warn("Attempting to start a Raptor Master without RP execution environment.")

    rp_version = packaging.version.parse(rp.version)

    args = parser.parse_args()
    if not os.path.exists(args.file):
        raise RuntimeError(f"File not found: {args.file}")
    with open(args.file, "r") as fh:
        configuration = MasterTaskConfiguration.from_dict(json.load(fh))

    logger.info(f"Launching ScaleMSMaster task in raptor {rp_version} with {repr(dataclasses.asdict(configuration))}.")

    _master = ScaleMSMaster(configuration)
    logger.debug(f"Created {repr(_master)}.")

    requirements = configuration.worker

    try:
        with _master.configure_worker(requirements=requirements) as configs:
            assert len(configs) == 1
            worker_submission: rp.TaskDescription = configs[0]
            assert os.path.isabs(worker_submission.raptor_file)
            assert os.path.exists(worker_submission.raptor_file)
            logger.debug(f"Submitting {len(configs)} {worker_submission.as_dict()}")
            _master.submit_workers(descriptions=configs)
            message = "Submitted workers: "
            message += ", ".join(_master.workers.keys())
            logger.info(message)

            _master.start()
            logger.debug("Master started.")

            # Make sure at least one worker comes online.
            _master.wait_workers(count=1)
            logger.debug("Ready to submit raptor tasks.")
            # Confirm all workers start successfully or produce useful error
            #  (then release temporary file).
            _master.wait_workers()
        for uid, worker in _master.workers.items():
            logger.info(f"Worker {uid} in state {worker['status']}.")
        _master.join()
        logger.debug("Master joined.")
    finally:
        if sys.exc_info() != (None, None, None):
            logger.exception("Master task encountered exception.")
        logger.debug("Completed master task.")


def _configure_worker(*, requirements: ClientWorkerRequirements, filename: str) -> RaptorWorkerConfig:
    """Prepare the arguments for :py:func:`rp.raptor.Master.submit_workers()`.

    Provide an abstraction for the submit_workers signature and parameter types,
    which may still be evolving.

    See also https://github.com/radical-cybertools/radical.pilot/issues/2731
    """
    assert os.path.exists(filename)
    # TODO(#248): Consider a "debug-mode" option to do a trial import from *filename*
    # TODO(#302): Number of workers should be chosen after inspecting the requirements of the work load.
    num_workers = 1
    descr = worker_description(
        named_env=requirements.named_env,
        worker_file=filename,
        pre_exec=requirements.pre_exec,
        cpu_processes=requirements.cpu_processes,
        gpus_per_process=requirements.gpus_per_process,
    )

    config: RaptorWorkerConfig = [descr] * num_workers
    return config


class SoftwareCompatibilityError(RuntimeError):
    """Incompatible package versions or software interfaces."""


def _get_module_version(module: str):
    """Get version metadata for importable module."""
    try:
        found_version = importlib.metadata.version(module)
    except importlib.metadata.PackageNotFoundError:
        found_version = None
    if found_version is None:
        try:
            spec: ModuleSpec = find_spec(module)
            if spec and hasattr(spec, "parent"):
                found_version = importlib.metadata.version(spec.parent)
        except Exception as e:
            logger.debug(f"Exception when trying to find {module}: ", exc_info=e)
    if found_version is not None:
        found_version = packaging.version.Version(found_version)
    return found_version


class ScaleMSMaster(rp.raptor.Master):
    """Manage a RP Raptor scheduler target.

    Extends :py:class:`rp.raptor.Master`.

    We do not submit scalems tasks directly as RP Tasks from the client.
    We encode scalems tasks as instructions to the Master task, which will be
    decoded and translated to RP Tasks by the `ScaleMSMaster.request_cb` and
    self-submitted to the `ScaleMSWorker`.

    Results of such tasks are only available through to the :py:func:`result_cb`.
    The Master can translate results of generated Tasks into results for the Task
    carrying the coded instruction, or it can produce data files to stage during or
    after the Master task. Additionally, the Master could respond to other special
    instructions (encoded in later client-originated Tasks) to query or retrieve
    generated task results.

    .. uml::

        title raptor Master

        queue "Queue" as Queue

        box "RP Agent"
        participant Scheduler
        end box

        queue "Master input queue" as master_queue

        box "scalems.radical.raptor"
        participant ScaleMSMaster
        participant rpt.Master
        end box

        queue "ZMQ Raptor work channel" as channel

        activate Queue

        Scheduler -> Queue: accepts work
        activate Scheduler
        Scheduler <-- Queue: task dictionary
        deactivate Queue

        note over Scheduler
        Scheduler gets task from queue,
        observes `scheduler` field and routes to Master.
        end note

        Scheduler -> master_queue: task dictionary
        activate master_queue
        deactivate Scheduler

        note over ScaleMSMaster, rpt.Master
        Back end RP queue manager
        passes messages to Master callbacks.
        end note

        rpt.Master -> master_queue: accept Task
        activate rpt.Master
        rpt.Master <-- master_queue: cpi_task
        deactivate master_queue

        rpt.Master -> rpt.Master: _request_cb(cpi_task)
        activate rpt.Master
        rpt.Master -> ScaleMSMaster: request_cb(cpi_task)
        activate ScaleMSMaster

        ScaleMSMaster -> ScaleMSMaster: CpiCommand.launch()
        activate ScaleMSMaster
        alt optionally process or update requests
          ScaleMSMaster -> ScaleMSMaster: submit_tasks(scalems_tasks)
          activate ScaleMSMaster
          deactivate ScaleMSMaster
          ScaleMSMaster -> ScaleMSMaster: _result_cb(cpi_task)
          activate ScaleMSMaster
          deactivate ScaleMSMaster
        else resolve a Control message
          ScaleMSMaster -> ScaleMSMaster: _result_cb(cpi_task)
          activate ScaleMSMaster
          deactivate ScaleMSMaster
        end
        return

        return filtered tasks

        rpt.Master -> rpt.Master: submit_tasks(...)
        activate rpt.Master
        deactivate rpt.Master
        rpt.Master -> channel: send message
        deactivate rpt.Master
        deactivate rpt.Master

        rpt.Master -> channel: accept result
        activate rpt.Master
        rpt.Master <-- channel: scalems_tasks
        deactivate channel

        rpt.Master -> rpt.Master: _result_cb(scalems_tasks))
        activate rpt.Master
        rpt.Master -> ScaleMSMaster: result_cb(scalems_tasks)
        activate ScaleMSMaster

        ScaleMSMaster -> ScaleMSMaster: CpiCommand.result_hook()

        alt optionally process or update pending requests
        ScaleMSMaster -> ScaleMSMaster: issue#277
        activate ScaleMSMaster
        deactivate ScaleMSMaster
        end

        rpt.Master <-- ScaleMSMaster
        deactivate ScaleMSMaster

        rpt.Master -> master_queue: send message
        activate master_queue
        deactivate rpt.Master
        deactivate rpt.Master

    """

    def __init__(self, configuration: MasterTaskConfiguration):
        """Initialize a SCALE-MS Raptor Master.

        Verify the environment. Perform some scalems-specific set up and
        initialize the base class details.
        """
        # Verify environment.
        for module, version in configuration.versioned_modules:
            logger.debug(f"Looking for {module} version {version}.")
            found_version = _get_module_version(module)
            if found_version is None:
                raise SoftwareCompatibilityError(f"{module} not found.")
            minimum_version = packaging.version.Version(version)
            if found_version >= minimum_version:
                logger.debug(f"Found {module} version {found_version}: Okay.")
            else:
                raise SoftwareCompatibilityError(f"{module} version {found_version} not compatible with {version}.")
        # The rp.raptor base class will fail to initialize without the environment
        # expected for a RP task launched by an RP agent. However, we may still
        # want to acquire a partially-initialized instance for testing.
        try:
            super(ScaleMSMaster, self).__init__()
        except KeyError:
            warnings.warn("Master incomplete. Could not initialize raptor.Master base class.")

        # Initialize internal state.
        # TODO: Use a scalems RuntimeManager.
        self.__worker_files = {}

    @contextlib.contextmanager
    def configure_worker(self, requirements: ClientWorkerRequirements) -> Generator[RaptorWorkerConfig, None, None]:
        """Scoped temporary module file for raptor worker.

        Write and return the path to a temporary Python module. The module imports
        :py:class:`ScaleMSWorker` into its module namespace
        so that the file and class can be used in the worker description for
        :py:func:`rp.raptor.Master.submit_workers()`
        """
        with self._worker_file() as worker_file:
            assert os.path.abspath(worker_file)
            assert os.path.exists(worker_file)
            worker_config = _configure_worker(requirements=requirements, filename=worker_file)
            yield worker_config

    @contextlib.contextmanager
    def _worker_file(self):
        """Scoped temporary module file for raptor worker.

        Not thread-safe. (Should it be?)

        Write and return the path to a temporary Python module. The module imports
        :py:class:`ScaleMSWorker` into its module namespace
        so that the file and class can be used in the worker description for
        :py:func:`rp.raptor.Master.submit_workers()`
        """
        worker = ScaleMSWorker
        # Note: there does not appear to be any utility in any lines to evaluate
        # beyond the Worker subclass import. This namespace will not be available
        # (through `globals()` or `locals()`, at least) to the rp.TASK_FUNCTION
        # implementation in rp.raptor.worker_mpi._Worker._dispatch_function.
        text = [
            f"from {worker.__module__} import {worker.__name__}\n",
            # f'from {run_in_worker.__module__} import {run_in_worker.__name__}\n'
        ]
        # TODO: Use the Master's FileStore to get an appropriate fast shared filesystem.
        with tempfile.TemporaryDirectory() as tmp_dir:
            filename = self.__worker_files.get(worker, None)
            if filename is None:
                filename = next_numbered_file(dir=tmp_dir, name="scalems_worker", suffix=".py")
                with open(filename, "w") as fh:
                    fh.writelines(text)
                self.__worker_files[worker] = filename
            yield filename
        del self.__worker_files[worker]

    def request_cb(self, tasks: typing.Sequence[TaskDictionary]) -> typing.Sequence[TaskDictionary]:
        """Allows all incoming requests to be processed by the Master.

        RADICAL guarantees that :py:func:`~radical.pilot.raptor.Master.request_cb()`
        calls are made sequentially in a single thread.
        The implementation does not need to be thread-safe or reentrant.
        (Ref: https://github.com/radical-cybertools/radical.utils/blob/master/src/radical/utils/zmq/queue.py#L386)

        RADICAL does not guarantee that Tasks are processed in the same order in
        which they are submitted by the client.

        If overridden, request_cb() must return a :py:class:`list`. The returned list
        is interpreted to be requests that should be processed normally after the callback.
        This allows subclasses of rp.raptor.Master to add or remove requests before they become Tasks.
        The returned list is submitted with self.submit_tasks() by the base class after the
        callback returns.

        A Master may call *self.submit_tasks()* to self-submit items (e.g. instead of or in
        addition to manipulating the returned list in *request_cb()*).

        It is the developer's responsibility to choose unique task IDs (uid) when crafting
        items for Master.submit_tasks().
        """
        try:
            # It is convenient to compartmentalize the filtering and dispatching
            # of scalems tasks in a generator function, but this callback signature
            # requires a `list` to be returned.
            remaining_tasks = self._scalems_handle_requests(tasks)
            # Let non-scalems tasks percolate to the default machinery.
            return list(remaining_tasks)
        except Exception as e:
            logger.exception("scalems request filter propagated an exception.")
            # TODO: Submit a clean-up task.
            #  Use a thread or asyncio event loop controlled by the main master task thread
            #  to issue a cancel to all outstanding tasks and stop the workers.
            # Question: Is there a better / more appropriate way to exit cleanly on errors?
            self.stop()
            # Note that we are probably being called in a non-root thread.
            # TODO: Avoid letting an exception escape unhandled.
            # WARNING: RP 1.18 suppresses exceptions from request_cb(), but there is a note
            # indicating that exceptions will cause task failure in a future version.
            # But the failure modes of the scalems master task are not yet well-defined.
            # See also
            # * https://github.com/SCALE-MS/scale-ms/issues/218 and
            # * https://github.com/SCALE-MS/scale-ms/issues/229
            raise e

    def cpi_finalize(self, task: TaskDictionary):
        """Short-circuit the normal Raptor protocol to finalize a task.

        This is an alias for Master._result_cb(), but is a public method used
        in the :py:class:`CpiCommand.launch` method for
        :py:class:`~scalems.messages.Control` commands, which do not
        call :py:func:`Master.submit_tasks`.
        """
        self._result_cb(task)

    def _scalems_handle_requests(self, tasks: typing.Sequence[TaskDictionary]):
        for task in tasks:
            _finalized = False
            try:
                mode = task["description"]["mode"]
                # Allow non-scalems work to be handled normally.
                if mode != CPI_MESSAGE:
                    logger.debug(f"Deferring {str(task)} to regular RP handling.")
                    yield task
                    continue
                else:
                    command = scalems.messages.Command.decode(task["description"]["metadata"])
                    logger.debug(f"Received message {command} in {str(task)}")

                    impl: typing.Type[CpiCommand] = CpiCommand.get(command)

                    # # TODO(#277)
                    # # Check work dependencies. Generate necessary data staging.
                    # self._dependents[task['uid']] = ...
                    # # Generate sub-tasks and manage dependencies. Prune tasks that
                    # # are already completed. Produce appropriate error if work cannot
                    # # be performed. Manage sufficient state that result_cb is able to
                    # # submit tasks as their dependencies are met.
                    #
                    # self._workload.add(task)
                    # ...

                    impl.launch(self, task)
                    # Note: this _finalized flag is immediately useless for non-Control commands.
                    # TODO: We need more task state maintenance, probably with thread-safety.
                    _finalized = True
            except Exception as e:
                # Exceptions here are presumably bugs in scalems or RP.
                # Almost certainly, we should shut down after cleaning up.
                # 1. TODO: Record error in log for ScaleMSMaster.
                # 2. Make sure that task is resolved.
                # 3. Trigger clean shut down of master task.
                if not _finalized:
                    # TODO: Mark failed, or note exception
                    self._result_cb(task)
                raise e

    def result_cb(self, tasks: typing.Sequence[TaskDictionary]):
        """SCALE-MS specific handling of completed tasks.

        We perform special handling for two types of tasks.
        1. Tasks submitted throughcompleted by the collaborating Worker(s).
        2. Tasks intercepted and resolved entirely by the Master (CPI calls).
        """
        # Note: At least as of RP 1.18, exceptions from result_cb() are suppressed.
        for task in tasks:
            logger.debug(f'Result callback for {task["description"]["mode"]} task {task["uid"]}: {task}')
            mode = task["description"]["mode"]
            # Allow non-scalems work to be handled normally.
            if mode == CPI_MESSAGE:
                command = scalems.messages.Command.decode(task["description"]["metadata"])
                logger.debug(f"Finalizing {command} in {str(task)}")

                impl: typing.Type[CpiCommand] = CpiCommand.get(command)
                impl.result_hook(self, task)

            # # Release dependent tasks.
            # #  https://github.com/SCALE-MS/randowtal/blob/c58452a3eaf0c058337a85f4fca3f51c5983a539/test_am
            # #  /brer_master.py#L114
            # # TODO: use scalems work graph management.
            # if task['uid'] in self._dependents:
            #     # Warning: The dependent task may have multiple dependencies.
            #     dep = self._dependents[task['uid']]
            #     self._log.debug('=== submit dep  %s', dep['uid'])
            #     self.submit_tasks(dep)
            #
            # mode = task['description']['mode']
            #
            # # NOTE: `state` will be `AGENT_EXECUTING`
            # self._log.debug('=== out: %s', task['stdout'])
            # self._log.debug('result_cb  %s: %s [%s] [%s]',
            #                 task['uid'],
            #                 task['state'],
            #                 sorted(task['stdout']),
            #                 task['return_value'])

            # #TODO(#108)
            # # Complete and publish
            # #  https://github.com/SCALE-MS/randowtal/blob/c58452a3eaf0c058337a85f4fca3f51c5983a539/test_am
            # #  /brer_master.py#L114
            # # Check whether all of the submitted tasks have been completed.
            # if self._submitted == self._completed:
            #     # we are done, not more workloads to wait for - return the tasks
            #     # and terminate
            #     # TODO: We can advance the rp task state with a callback registered with the corresponding
            #     #  unwrapped scalems task.
            #     for req in self._workload:
            #         # TODO: create raptor method
            #         req['task']['target_state'] = rp.DONE
            #         self.advance(req['task'], rp.AGENT_STAGING_OUTPUT_PENDING,
            #                      publish=True, push=True)


class WorkerDescription(rp.TaskDescription):
    """Worker description.

    See Also:
        * :py:meth:`~radical.pilot.raptor.Master.submit_workers()`
        * https://github.com/radical-cybertools/radical.pilot/issues/2731

    """

    cores_per_rank: typing.Optional[int]
    environment: typing.Optional[dict[str, str]]
    gpus_per_rank: typing.Optional[int]
    named_env: typing.Optional[str]
    pre_exec: typing.Optional[list[str]]
    ranks: int
    raptor_class: str
    raptor_file: str
    mode: str  # rp.RAPTOR_WORKER


RaptorWorkerConfig = list[WorkerDescription]
"""Client-specified Worker requirements.

Used by master script when calling `worker_description()`.
Created internally with :py:func:`_configure_worker()`.

See Also:
    :py:mod:`scalems.radical.raptor.ClientWorkerRequirements`

"""


def worker_description(
    *,
    named_env: str,
    worker_file: str,
    cores_per_process: int = None,
    cpu_processes: int = None,
    gpus_per_process: int = None,
    pre_exec: typing.Iterable[str] = (),
    environment: typing.Optional[typing.Mapping[str, str]] = None,
) -> WorkerDescription:
    """Get a worker description for Master.submit_workers().

    Parameters:
        cores_per_process (int, optional): See `radical.pilot.TaskDescription.cores_per_rank`
        cpu_processes (int, optional): See `radical.pilot.TaskDescription.ranks`
        environment (dict, optional): Environment variables to set in the Worker task.
        gpus_per_process (int, optional): See `radical.pilot.TaskDescription.gpus_per_rank`
        named_env (str): Python virtual environment registered with `radical.pilot.Pilot.prepare_env`
            (currently ignored. see #90).
        pre_exec (list[str]): Shell command lines for preparing the worker environment.
        worker_file (str): Standalone module from which to import ScaleMSWorker.

    *worker_file* is generated for the master script by the caller.
    See :py:mod:`scalems.radical.raptor`

    The *uid* for the Worker task is defined by the Master.submit_workers().
    """
    descr = WorkerDescription()
    if cores_per_process is not None:
        descr.cores_per_rank = cores_per_process
    if environment is not None:
        descr.environment = environment
    if gpus_per_process is not None:
        descr.gpus_per_rank = gpus_per_process
    if named_env is not None:
        # descr.named_env=None
        logger.debug(f"Ignoring named_env={named_env}. Parameter is currently unused.")
    descr.pre_exec = list(pre_exec)
    if cpu_processes is not None:
        descr.ranks = cpu_processes
    descr.raptor_class = "ScaleMSWorker"
    descr.raptor_file = worker_file
    descr.mode = rp.RAPTOR_WORKER
    return descr


class RaptorTaskExecutor(typing.Protocol):
    """Represent the signature of executor functions for rp.raptor.MPIWorker."""

    def __call__(self, *args, comm: Comm, **kwargs) -> Encodable:
        ...


class ScaleMSWorker(rp.raptor.MPIWorker):
    """Specialize the Raptor MPI Worker for scalems dispatching of serialised work.

    scalems tasks encode the importable function and inputs in the arguments to
    the `run_in_worker` dispatching function,
    which is available to the `ScaleMSWorker` instance.

    .. uml::

        title ScaleMSWorker task dispatching

        queue "ZMQ Raptor work channel" as channel

        box "scalems.radical.raptor.Worker"
        participant ScaleMSWorker
        participant rpt.Worker
        end box

        box "usermodule"
        participant usermodule
        end box

        box "target venv"
        end box

        rpt.Worker -> channel: pop message
        activate rpt.Worker
        rpt.Worker <-- channel: rp.TASK_FUNCTION mode Task
        rpt.Worker -> rpt.Worker: _dispatch_function()
        activate rpt.Worker

        alt scalems encoded workload
          rpt.Worker -> ScaleMSWorker: run_in_worker()
          activate ScaleMSWorker
          ScaleMSWorker -> ScaleMSWorker: unpack encoded function call
          ScaleMSWorker -> ScaleMSWorker: from usermodule import func
          ScaleMSWorker -> usermodule: ""func(*args, **kwargs)""
          activate usermodule
          ScaleMSWorker <-- usermodule
          deactivate usermodule
          return
        else pickled rp.PythonTask workload (not implemented)
          rpt.Worker -> rpt.Worker: unpickled scalems handler
          activate rpt.Worker
          rpt.Worker -> usermodule: ""func(*args, **kwargs)""
          activate usermodule
          rpt.Worker <-- usermodule
          deactivate usermodule
          rpt.Worker --> rpt.Worker: {out, err, ret, value}
          deactivate rpt.Worker
        end

        deactivate rpt.Worker
        rpt.Worker -> channel: put result
        deactivate rpt.Worker

    """

    def run_in_worker(self, *, work_item: ScalemsRaptorWorkItem, comm=None):
        """Unpack and run a task requested through RP Raptor.

        Satisfies RaptorTaskExecutor protocol. Named as the *function* for
        rp.TASK_FUNCTION mode tasks generated by scalems.

        This function MUST be imported and referentiable by its *__qualname__* from
        the scope in which ScaleMSWorker methods execute. (See ScaleMSMaster._worker_file)

        Parameters:
            comm (mpi4py.MPI.Comm, optional): MPI communicator to be used for the task.
            work_item: dictionary of encoded function call from *kwargs* in the TaskDescription.

        See Also:
            :py:func:`CpiAddItem.launch()`
        """
        module = importlib.import_module(work_item["module"])
        func = getattr(module, work_item["func"])
        args = list(work_item["args"])
        kwargs = work_item["kwargs"].copy()
        comm_arg_name = work_item.get("comm_arg_name", None)
        if comm_arg_name is not None:
            if comm_arg_name:
                kwargs[comm_arg_name] = comm
            else:
                args.append(comm)
        logger.debug(
            "Calling {func} with args {args} and kwargs {kwargs}",
            {"func": func.__qualname__, "args": repr(args), "kwargs": repr(kwargs)},
        )
        return func(*args, **kwargs)


def raptor_work_deserializer(*args, **kwargs):
    """Unpack a workflow document from a Task.

    When a `ScaleMSMaster.request_cb()` receives a raptor task, it uses this
    function to unpack the instructions and construct a work load.
    """
    raise scalems.exceptions.MissingImplementationError


class ScalemsRaptorWorkItem(typing.TypedDict):
    """Encode the function call to implement a task in the workflow.

    Parameter type for the `run_in_worker` function *work_item* argument.

    This structure must be trivially serializable (by `msgpack`) so that it can
    be passed in a TaskDescription.kwargs field. Consistency with higher level
    `scalems.serialization` is not essential, but can be pursued in the future.

    The essential role is to represent an importable callable and its arguments.
    At run time, the dispatching function (`run_in_worker`) conceivably has access
    to module scoped state, but does not have direct access to the Worker (or
    any resources other than the `mpi4py.MPI.Comm` object). Therefore, abstract
    references to input data should be resolved at the Master in terms of the
    expected Worker environment before preparing and submitting the Task.

    TODO: Evolve this to something sufficiently general to be a scalems.WorkItem.

    TODO: Clarify the translation from an abstract representation of work in terms
     of the workflow record to the concrete representation with literal values and
     local filesystem paths.

    TODO: Record extra details like implicit filesystem interactions,
     environment variables, etc. TBD whether they belong in a separate object.
    """

    func: str
    """A callable to be retrieved as an attribute in *module*."""

    module: str
    """The qualified name of a module importable by the Worker."""

    args: list
    """Positional arguments for *func*."""

    kwargs: dict
    """Key word arguments for *func*."""

    comm_arg_name: typing.Optional[str]
    """Identify how to provide an MPI communicator to *func*, if at all.

    If *comm_arg_name* is not None, the callable will be provided with the
    MPI communicator. If *comm_arg_name* is an empty string, the communicator
    is provided as the first positional argument. Otherwise, *comm_arg_name*
    must be a valid key word argument by which to pass the communicator to *func*.
    """


class _RaptorTaskDescription(typing.TypedDict):
    """Describe a Task to be executed through a Raptor Worker.

    A specialization of `radical.pilot.TaskDescription`.

    Note the distinctions of a TaskDescription to be processed by a raptor.Master.

    The meaning or utility of some fields is dependent on the values of other fields.

    TODO: We need some additional fields, like *environment* and fields related
     to launch method and resources reservation. Is the schema for rp.TaskDescription
     sufficiently strong and well-documented that we can remove this hinting type?
     (Check again at RP >= 1.19)
    """

    uid: str
    """Unique identifier for the Task across the Session."""

    executable: str
    """Unused by Raptor tasks."""

    scheduler: str
    """The UID of the raptor.Master scheduler task.

    This field is relevant to tasks routed from client TaskManagers. It is not
    used for tasks originating in master tasks.

    .. ref https://github.com/SCALE-MS/scale-ms/discussions/258#discussioncomment-4087870

    """

    metadata: Encodable
    """An arbitrary user-provided payload.

    May be any type that is encodable by :py:mod:`msgpack` (i.e. built-in types).
    """

    mode: str
    """The executor mode for the Worker to use.

    For ``rp.TASK_FUNCTION`` mode, either *function* or *method* must name a task executor.
    Depending on the Worker (sub)class, resources such as an `mpi4py.MPI.Comm`
    will be provided as the first positional argument to the executor.
    ``*args`` and ``**kwargs`` will be provided to the executor from the corresponding
    fields.

    See Also:
        * :py:data:`CPI_MESSAGE`
        * :py:class:`RaptorTaskExecutor`
    """

    function: str
    """Executor for ``rp.TASK_FUNCTION`` mode.

    Names the callable for dispatching.

    The callable can either be a function object present in the namespace of the interpreter launched
    by the Worker for the task, or a :py:class:`radical.pilot.pytask.PythonTask` pickled function object.
    """

    args: list
    """For ``rp.TASK_FUNCTION`` mode, list of positional arguments provided to the executor function."""

    kwargs: dict
    """For ``rp.TASK_FUNCTION`` mode, a dictionary of key word arguments provided to the executor function."""


class TaskDictionary(typing.TypedDict):
    """Task representations seen by *request_cb* and *result_cb*.

    Other fields may be present, but the objects in the sequences provided to
    :py:meth:`scalems.radical.raptor.ScaleMSMaster.request_cb()` and
    :py:meth:`scalems.radical.raptor.ScaleMSMaster.result_cb()` have the following fields.
    Result fields will not be populated until the Task runs.

    For the expected fields, see the source code for
    :py:meth:`~radical.pilot.Task.as_dict()`:
    https://radicalpilot.readthedocs.io/en/stable/_modules/radical/pilot/task.html#Task.as_dict
    """

    uid: str
    """Canonical identifier for the Task.

    Note that *uid* may be omitted from the original TaskDescription.
    """

    description: _RaptorTaskDescription
    """Encoding of the original task description."""

    stdout: str
    """Task standard output."""

    stderr: str
    """Task standard error."""

    exit_code: int
    """Task return code."""

    return_value: typing.Any
    """Function return value.

    Refer to the :py:class:`RaptorTaskExecutor` Protocol.
    """

    exception: str
    """Exception type name and message."""

    exception_detail: str
    """Full exception details with stack trace."""

    state: str
    """RADICAL Pilot Task state."""

    target_state: str
    """State to which the Task should be advanced.

    Valid values are string constants from :py:mod:`radical.pilot.states`.

    Used internally, such as in Master._result_cb().
    """


def next_numbered_file(*, dir, name, suffix):
    """Get the next available filename with the form {dir}/{name}{i}{suffix}."""
    if not os.path.exists(dir):
        raise ValueError(f'Directory "{dir}" does not exist.')
    filelist = os.listdir(dir)
    template = str(name) + "{i}" + str(suffix)
    i = 1
    filename = template.format(i=i)
    while filename in filelist:
        i += 1
        filename = template.format(i=i)
    return os.path.join(dir, filename)



"""Support command line execution of scalems packaged function calls.

Example:
    python -m scalems.call record.json

Note that this whole module is a workaround to avoid using RP raptor.
:file:`scalems/call/__main__.py` provides command line behavior for a
`radical.pilot.Task` *executable* and *arguments*, with which we execute a
serialized representation of a function call, pending restoration of full
"raptor" support.

We don't really need to reconcile the divergent notions of "Subprocess",
if we can remove the need for this entry point.
But we _do_ need to formalize our serialized representation of a function call,
and the task lifecycle and artifacts management.
"""
__all__ = ("CallResult", "main", "cli", "serialize_call", "deserialize_result")

import dataclasses
import functools
import json

# See __main__.py for the entry point executed for command line invocation.

import logging
import os
import pathlib
import sys
import tempfile
import typing
import warnings

import dill

import scalems.store as _store

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@dataclasses.dataclass
class _Call:
    """Record of a function call to be dispatched."""

    func: typing.Callable
    """Python callable.

    Assumed to be round-trip serializable via *encoder* and *decoder*.
    """

    environment: typing.Optional[dict] = dataclasses.field(default_factory=dict)
    """Optional overrides of environment variables."""

    skeleton: typing.Optional[str] = None
    """Optional string-encoded URI for archived skeleton of working directory files."""

    args: typing.Sequence = dataclasses.field(default_factory=tuple)
    """Positional arguments.

    Values are assumed to be round-trip serializable via *encoder* and *decoder*.
    """

    kwargs: dict = dataclasses.field(default_factory=dict)
    """Dictionary of key word arguments and values.

    Values are assumed to be round-trip serializable via *encoder* and *decoder*.
    """

    outputs: list[str] = dataclasses.field(default_factory=list)
    """Requested outputs.

    To avoid excessive data transfer, outputs (ResultPack fields)
    must be named explicitly to be included in the automatically
    generated result package.
    """

    requirements: dict[str, typing.Any] = dataclasses.field(default_factory=dict)
    """Named run time requirements for the call.

    Common requirements include numbers of MPI ranks, cores per process, or
    descriptions of accelerator hardware (GPUs).

    For the initial implementation, key/value pairs are assumed to map to
    attribute assignments in a `radical.pilot.TaskDescription`. The information
    is not provided directly to the function; it is for internal use.
    (Special handling may occur in `scalems.call`.)
    """

    decoder: str = dataclasses.field(default="dill")
    encoder: str = dataclasses.field(default="dill")
    # TODO: Describe return value type.


@dataclasses.dataclass
class CallResult:
    """Result type (container) for a dispatched function call."""

    return_value: typing.Optional[typing.Any] = None
    """Return value, if any.

    Value assumed to be round-trip serializable via *encoder* and *decoder*.
    """

    exception: typing.Optional[str] = None
    """string representation of the Exception, if any."""

    # Side effects
    stdout: typing.Optional[str] = None
    """string-encoded URI for file holding captured stdout."""

    stderr: typing.Optional[str] = None
    """string-encoded URI for file holding captured stderr."""

    directory: typing.Optional[str] = None
    """string-encoded URI for archive of the working directory after task execution."""

    # support
    decoder: str = dataclasses.field(default="dill")
    encoder: str = dataclasses.field(default="dill")


class CallPack(typing.TypedDict):
    """Trivially serializable representation of `scalems.call.Call`.

    Pack and unpack with `scalems.pack.pack_call` and `scalems.pack.unpack_call`.

    Serialize and deserialize with::

        serialized = json.dumps(dataclasses.asdict(callpack), separators=(",", ":"))
        callpack = CallPack(*json.loads(serialized))

    """

    args: list[str]
    """List of serialized positional arguments."""

    environment: typing.Optional[dict]
    """Optional overrides of environment variables."""

    func: str
    """string-encoded Python callable, deserializable with *decoder*.

    If *decoder* consumes *bytes* instead of *str*, *callable* is interpreted
    as hexadecimal-encoded big-endian bytes.
    """

    kwargs: dict[str, str]
    """Dictionary of key word arguments and serialized values."""

    outputs: list[str]
    """Requested outputs.

    To avoid excessive data transfer, outputs (ResultPack fields other than
    *exception*, *encoder*, and *decoder*)
    must be named explicitly to be included in the automatically
    generated result package.
    """

    requirements: dict[str, typing.Any]
    """Run time requirements.

    This is a copy of the run time requirements submitted with the command. It
    is not directly available to the wrapped function, but its information may
    be used by the `scalems.call.main` implementation.

    See Also:
        https://github.com/SCALE-MS/scale-ms/discussions/302
    """

    skeleton: typing.Optional[str]
    """Optional string-encoded URI for archived skeleton of working directory files.

    Warning:
        With or without this file list, additional files will be created to support
        task execution. If this is a problem, we might be able to customize the RP
        task launch script to execute in a subdirectory.
    """

    decoder: str
    encoder: str


class ResultPack(typing.TypedDict):
    """Trivially serializable representation of `scalems.call.CallResult`.

    Serialize and deserialize with::

        serialized = json.dumps(dataclasses.asdict(resultpack), separators=(",", ":"))
        resultpack = CallPack(*json.loads(serialized))

    """

    return_value: typing.Optional[str]
    """string-encoded return value, if any."""

    exception: typing.Optional[str]
    """string representation of the Exception, if any."""

    # Side effects
    stdout: typing.Optional[str]
    """string-encoded URI for file holding captured stdout."""

    stderr: typing.Optional[str]
    """string-encoded URI for file holding captured stderr."""

    directory: typing.Optional[str]
    """string-encoded URI for the working directory detected during task execution."""

    # support
    decoder: str
    encoder: str


@dataclasses.dataclass(frozen=True)
class _Subprocess:
    """Simplified Subprocess representation.

    This exists to support initial implementation and testing of function_call_to_subprocess().
    To be reconciled with generalized Subprocess class after testing and feedback.
    """

    # Note: we can/should enforce uniqueness semantics.
    uid: str
    input_filenames: typing.Mapping[str, _store.FileReference]
    output_filenames: tuple
    executable: str
    arguments: tuple[str, ...]
    requirements: dict = dataclasses.field(default_factory=dict)


async def function_call_to_subprocess(
    func: typing.Callable, *, label: str, args: tuple = (), kwargs: dict = None, manager, requirements: dict = None
) -> _Subprocess:
    """
    Wrap a function call in a command line based on `scalems.call`.

    Args:
        func: A callable Python object.
        label: Name (prefix) for identifying tasks and artifacts.
        args: Positional arguments for *func*.
        kwargs: Key word arguments for *func*.
        manager: workflow manager instance.
        requirements: run time requirements (passed through to execution backend).

    Returns:
        Serializable representation of the executable task.

    Warnings:
        This is a temporary utility while we explore use cases and prepare to migrate
        back to :py:mod:`radical.pilot.raptor`. A user-facing tool should return a
        view on a workflow item, whereas this function produces even lower-level details.

    See Also:
        * :py:func:`scalems.radical.runtime.subprocess_to_rp_task()`
        * https://github.com/SCALE-MS/scale-ms/discussions/302

    """
    if kwargs is None:
        kwargs = dict()
    if requirements is None:
        requirements = dict()

    # TODO: FileStore should probably provide a new_file() method.
    #   Possibly an overload to get_file_reference() for IOStreams or buffers,
    #   but requires internal support to optimize out redundant fingerprinting
    #   and filesystem operations. I.e. it should be one read, one fingerprint
    #   calculation, one write, and one rename based on the fingerprint. The
    #   support could be in the form of `add_text`, and `add_blob` methods.
    #   Optimization for known binary formats (with potentially local differences)
    #   could be in the form of `add_typed_data`.
    with tempfile.NamedTemporaryFile(mode="w", suffix="-input.json") as tmp_file:
        tmp_file.write(serialize_call(func=func, args=args, kwargs=kwargs, requirements=requirements))
        tmp_file.flush()
        # We can't release the temporary file until the file reference is obtained.
        file_ref = await _store.get_file_reference(pathlib.Path(tmp_file.name), filestore=manager.datastore())

    uid = str(label)
    input_filename = uid + "-input.json"
    # TODO: Collaborate with scalems.call to agree on output filename.
    output_filename = uid + "-output.json"
    executable = "python3"
    arguments = []
    for key, value in getattr(sys, "_xoptions", {}).items():
        if value is True:
            arguments.append(f"-X{key}")
        else:
            assert isinstance(value, str)
            arguments.append(f"-X{key}={value}")

    # TODO: Validate with argparse, once scalems.call.__main__ has a real parser.
    if "ranks" in requirements:
        arguments.extend(("-m", "mpi4py"))
    # If this wrapper doesn't go away soon, we should abstract this so path arguments
    # can be generated at the execution site.
    arguments.extend(("-m", "scalems.call", input_filename, output_filename))

    return _Subprocess(
        uid=uid,
        input_filenames={input_filename: file_ref},
        output_filenames=(output_filename,),
        executable=executable,
        arguments=tuple(arguments),
        requirements=requirements.copy(),
    )


def main(call: _Call) -> CallResult:
    """Execute the packaged call.

    Return a packaged result.
    """
    func = call.func
    args = call.args
    kwargs = call.kwargs
    outputs = [str(output) for output in call.outputs]
    fields = [field.name for field in dataclasses.fields(CallResult)]
    for output in outputs:
        if output not in fields:
            # TODO: Make sure that warnings and logs go somewhere under RP.
            warnings.warn(f"Unrecognized output requested: {output}")
    if "exception" not in outputs:
        logger.debug("Adding *exception* to outputs for internal use.")
        outputs.append("exception")

    # Note: we are already running inside a RP executable Task by the time we
    # reach this point, so we are relying on call.environment and call.skeleton
    # to have been handled by the caller.
    cwd = pathlib.Path(os.getcwd())
    logger.info(f"scalems.call executing {func} in working directory {cwd}")
    result_fields = dict(directory=cwd.as_uri())

    # Note: For RP Raptor, the MPIWorker will be providing our wrapped function
    # with an mpi4py.MPI.Comm according to the task requirements. For this CLI
    # utility, RP will have used an appropriate launch method, so the parallelization
    # libraries should be able to initialize themselves from the environment.
    # TODO: Check that *e.g. gromacs* properly detects resources as prepared by RP.
    try:
        result_fields["return_value"] = func(*args, **kwargs)
    except Exception as e:
        result_fields["exception"] = repr(e)

    # Note: We don't have *stdout* and *stderr* at this point because
    # they are handled by the work load manager.
    kwargs = {key: value for key, value in result_fields.items() if key in outputs}
    result = CallResult(**kwargs)
    return result


def cli(*argv: str):
    """Command line entry point.

    Invoke with ``python -m scalems.call <args>``

    TODO: Configurable log level.
    """
    logger.setLevel(logging.DEBUG)
    character_stream = logging.StreamHandler()
    character_stream.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    character_stream.setFormatter(formatter)
    logger.addHandler(character_stream)

    file_logger = logging.FileHandler("scalems.call.log")
    file_logger.setLevel(logging.DEBUG)
    file_logger.setFormatter(formatter)
    logging.getLogger("scalems").addHandler(file_logger)

    logger.debug(f"scalems.call got args: {', '.join(str(arg) for arg in argv)}")
    # TODO: Consider an argparse parser for clarity.
    if len(argv) < 3:
        raise RuntimeError("Arguments are required.")
    # argv[0] will be the __main__.py script. Arguments to `call` are at argv[1:]
    call_path = pathlib.Path(argv[1])
    result_path = pathlib.Path(argv[2])
    with open(call_path, "r") as fh:
        call: _Call = deserialize_call(fh.read())
    result: CallResult = main(call)
    with open(result_path, "w") as fh:
        fh.write(serialize_result(result))
    if result.exception is not None:
        logging.error(result.exception)
        return 1
    else:
        return 0


# For transfer size and debuggability, let's start by serializing as little
# as possible. We'll rely on being able to import details in the execution
# environment (accepting whatever version we get locally) and work from there.
to_bytes = functools.partial(dill.dumps, byref=True, recurse=False)


def from_hex(x: str):
    return dill.loads(bytes.fromhex(x))


def serialize_call(func: typing.Callable, *, args: tuple = (), kwargs: dict = None, requirements: dict = None) -> str:
    """Create a serialized representation of a function call.

    This utility function is provided for stability while the serialization
    machinery and the CallPack structure evolve.
    """
    if kwargs is None:
        kwargs = {}
    if requirements is None:
        requirements = {}
    # This might need to be replaced with a dispatching function, for which raw serialization
    # is only a fall-back for functions that aren't decorated/registered scalems Commands.
    # Commands could be implemented for specific Runtime executors, or expressed in terms
    # of importable callables with similarly expressable Input and Result types.
    # Ref: https://github.com/SCALE-MS/scale-ms/issues/33
    serialized_callable: bytes = to_bytes(func)
    pack = CallPack(
        func=serialized_callable.hex(),
        args=[to_bytes(arg).hex() for arg in args],
        kwargs={key: to_bytes(value).hex() for key, value in kwargs.items()},
        decoder="dill",
        encoder="dill",
        outputs=["return_value", "exception", "directory", "stdout", "stderr"],
        skeleton=None,
        environment=None,
        requirements=requirements,
    )
    return json.dumps(pack, separators=(",", ":"))


def deserialize_call(record: str) -> _Call:
    """Deserialize a function call."""
    record_dict: CallPack = json.loads(record)
    assert record_dict["encoder"] == "dill"
    assert record_dict["decoder"] == "dill"
    # TODO: Pre-load references.
    func = dill.loads(bytes.fromhex(record_dict["func"]))
    args = tuple(from_hex(arg) for arg in record_dict["args"])
    kwargs = {key: from_hex(value) for key, value in record_dict["kwargs"].items()}
    outputs = record_dict["outputs"]
    call = _Call(
        func=func,
        args=args,
        kwargs=kwargs,
        outputs=outputs,
        environment=record_dict.get("environment", None),
        skeleton=record_dict.get("skeleton", None),
        requirements=record_dict.get("requirements", None),
    )
    return call


def serialize_result(result: CallResult) -> str:
    assert result.encoder == "dill"
    assert result.decoder == "dill"
    if result.return_value is not None:
        value = to_bytes(result.return_value).hex()
    else:
        value = None
    if result.exception is not None:
        exception = to_bytes(result.exception).hex()
    else:
        exception = None

    # In scalems.call.cli, we are not managing stdout and stderr redirection.
    # If the caller is managing stdout and stderr, the caller can repackage the
    # result with appropriate URIs.
    pack = ResultPack(
        return_value=value,
        exception=exception,
        stderr=None,
        stdout=None,
        directory=result.directory,
        encoder="dill",
        decoder="dill",
    )
    return json.dumps(pack, separators=(",", ":"))


def deserialize_result(stream: str) -> CallResult:
    pack: ResultPack = json.loads(stream)
    assert pack["encoder"] == "dill"
    assert pack["decoder"] == "dill"

    value = pack.get("return_value", None)
    if value is not None:
        assert isinstance(value, str)
        value = from_hex(value)
    exception = pack.get("exception", None)
    if exception is not None:
        assert isinstance(exception, str)
        exception = from_hex(exception)
    result = CallResult(
        return_value=value,
        exception=exception,
        stdout=pack.get("stdout", None),
        stderr=pack.get("stderr", None),
        directory=pack.get("directory", None),
        decoder=pack["decoder"],
        encoder=pack["encoder"],
    )
    return result



"""SCALE-MS - Scalable Adaptive Large Ensembles of Molecular Simulations.

This package provides Python driven data flow scripting and graph execution
for molecular science computational research protocols.

Documentation is published online at https://scale-ms.readthedocs.io/.

Refer to https://github.com/SCALE-MS/scale-ms/wiki for development documentation.

Invocation:
    ScaleMS scripts describe a workflow. To run the workflow, you must tell
    ScaleMS how to dispatch the work for execution.

    For most use cases, you can run the script in the context of a particular
    execution scheme by using the ``-m`` Python command line flag to specify a
    ScaleMS execution module::

        # Execute with the default local execution manager.
        python -m scalems.local myworkflow.py
        # Execute with the RADICAL Pilot based execution manager.
        python -m scalems.radical myworkflow.py

    Execution managers can be configured and used from within the workflow script,
    but the user has extra responsibility to properly shut down the execution
    manager, and the resulting workflow may be less portable. For details, refer
    to the documentation for particular WorkflowContexts.

"""

# Note: Even though `from scalems import *` is generally discouraged, the __all__ module attribute is useful
# to document the intended public interface *and* to indicate sort order for tools like
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#directive-automodule
__all__ = [
    # core UI
    "app",
    # tools / commands
    "executable",
    # utilities and helpers
    "get_scope",
    # 'run',
    # 'wait',
    "__version__",
    # core API
    "ScriptEntryPoint",
]

import abc
import functools
import logging
import typing

from ._version import __version__
from .subprocess import executable
from .workflow import get_scope

logger = logging.getLogger(__name__)
logger.debug("Imported {}".format(__name__))


class ScriptEntryPoint(abc.ABC):
    """Annotate a SCALE-MS entry point function.

    An importable Python script may decorate a callable with scalems.app to
    mark it for execution. This abstract base class provides SCALE-MS with a
    way to identify callables marked for execution and is not intended to be
    used directly.

    See :py:func:`scalems.app`
    """

    name: typing.Optional[str]

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        ...


def app(func: typing.Callable) -> typing.Callable:
    """Annotate a callable for execution by SCALEMS."""

    class App(ScriptEntryPoint):
        def __init__(self, func: typing.Callable):
            if not callable(func):
                raise ValueError("Needs a function or function object.")
            self._callable = func
            self.name = None

        def __call__(self, *args, **kwargs):
            return self._callable(*args, **kwargs)

    decorated = functools.update_wrapper(App(func), wrapped=func)

    return decorated



"""Core typing information.

This module is intended to have no dependencies. It allows some core definitions to be
used in multiple modules without tightly coupling those modules.
"""
import typing

json_base_encodable_types: typing.Tuple[type, ...] = (dict, list, tuple, str, int, float, bool, type(None))

json_base_decoded_types: typing.Tuple[type, ...] = (dict, list, str, int, float, bool, type(None))
BaseEncodable = typing.Union[dict, list, tuple, str, int, float, bool, None]
BaseDecoded = typing.Union[dict, list, str, int, float, bool, None]



"""Test the scalems.executable command.

TODO: Consider whether we can reimplement as a parameterized test over the
      various executors.
"""

import asyncio
import logging

import pytest

import scalems.context
import scalems.local
import scalems.workflow
from scalems.subprocess import executable


@pytest.mark.asyncio
async def test_executable_local(cleandir):
    # For a similar test dispatching through RADICAL Pilot, see test_rp_exec.py::test_exec_rp
    asyncio.get_event_loop().set_debug(True)
    logging.getLogger("asyncio").setLevel(logging.DEBUG)

    # Test local execution with standard deferred launch.
    # Exercise 1:
    #   * Define the task before entering the dispatch context
    #   * Check the handling for stdin pipelines
    # Note that a coroutine object created from an `async def` function is only awaitable once.
    with scalems.workflow.scope(scalems.local.workflow_manager(asyncio.get_event_loop())) as context:
        # TODO: The `with` block should be equivalent to a `-m scalems.local` invocation. Test.
        # TODO: Input type checking.
        try:
            cmd = executable(("/bin/cat", "-"), stdin=("hi there\n", "hello world"), stdout="stdout.txt")
        except Exception as e:
            # This is just here to give us a place to set a "break" in a debugger.
            raise e
        assert isinstance(cmd, scalems.workflow.ItemView)
        # TODO: Future interface allows client to force resolution of dependencies.
        # cmd.result()
        # TODO(#82): Enable scalems.run()
        # scalems.run(cmd)
        # TODO: Remove Session.run() from public interface (use scalems.run())
        # await context.run()
        async with context.dispatch():
            ...
        result = cmd.result()  # type: scalems.subprocess.SubprocessResult
        assert result.stdout.name == "stdout.txt"
        with open(result.stdout) as fh:
            assert fh.read().startswith("hi there")
    # We need to close the filestore before we begin to initialize a new
    # WorkflowManager so that we don't try to take over an actively managed filestore.
    context.close()

    # Exercise 2:
    #   * Define the task while inside the dispatching context
    with scalems.workflow.scope(scalems.local.workflow_manager(asyncio.get_event_loop())) as context:
        # TODO: Future interface allows client to force resolution of dependencies.
        # cmd.result()
        # TODO(#82): Enable scalems.run()
        # scalems.run(cmd)
        async with context.dispatch():
            # TODO: Input type checking.
            cmd = executable(("/bin/echo", "hello", "world"), stdout="stdout.txt")
            assert isinstance(cmd, scalems.workflow.ItemView)
        result = cmd.result()  # type: scalems.subprocess.SubprocessResult
        assert result.stdout.name == "stdout.txt"
        with open(result.stdout) as fh:
            assert fh.read().startswith("hello world")
    # We need to close the filestore before removing the temporary directory to avoid
    # errors.
    context.close()



"""Define the ScaleMS Subprocess command.

`scalems.executable` is used to execute a program in one (or more) subprocesses.
It is an alternative to the built-in Python `subprocess.Popen` or `asyncio.create_subprocess_exec`
with extensions to better support ScaleMS execution dispatching and ensemble data flow.

The core task is represented by `scalems.subprocess.SubprocessTask`, but also
requires the definition of `scalems.subprocess.SubprocessInput` and
`scalems.subprocess.SubprocessResult`.



In the first iteration, we can use dataclasses.dataclass to define input/output data structures
in terms of standard types. In a follow-up, we can use a scalems metaclass to define them
in terms of Data Descriptors that support mixed scalems.Future and native constant data types.
"""

__all__ = ["executable", "Subprocess", "SubprocessInput", "SubprocessResult"]

import dataclasses
import json
import logging
import typing
from pathlib import Path  # We probably need a scalems abstraction for Path.

import scalems.workflow
from scalems.exceptions import InternalError
from scalems.serialization import encode
from ..unique import next_monotonic_integer
from ..exceptions import APIError
from ..exceptions import MissingImplementationError
from ..workflow import WorkflowManager

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))


class OutputFile(dict):
    """Placeholder for output files.

    The initial implementation of OutputFile does not provide access to the
    created output file.

    The actual filename is managed by SCALE-MS to avoid namespace collisions.

    Arguments:
        label (str): Optional user-friendly identifier for locating a reference in the managed workflow.
        suffix (str): Optional filename suffix for the generated filename.

    In a future implementation, we may allow instances of OutputFile to transform
    into workflow references that are dependent on the task under construction.
    """

    def __init__(self, label=None, suffix=""):
        super().__init__()
        self["label"] = label
        self["suffix"] = suffix

    @property
    def label(self):
        return self.get("label", None)

    @property
    def suffix(self):
        return self.get("suffix", "")


# TODO: what is the mechanism for registering a command implementation in a new Context?
# TODO: What is the relationship between the command factory and the command type? Which parts need to be importable?


# TODO: input data typing.
@dataclasses.dataclass(frozen=True)
class SubprocessInput:
    # TODO: Move input documentation to Input class docs.
    argv: typing.Sequence[str]
    inputs: typing.Mapping[str, Path] = dataclasses.field(default_factory=dict)
    outputs: typing.Mapping[str, OutputFile] = dataclasses.field(default_factory=dict)
    stdin: typing.Iterable[str] = ()
    environment: typing.Mapping[str, typing.Union[str, None]] = dataclasses.field(default_factory=dict)
    # For now, let's just always enable stdout/stderr
    stdout: Path = dataclasses.field(default=Path("stdout"))
    stderr: Path = dataclasses.field(default=Path("stderr"))
    resources: typing.Mapping[str, typing.Any] = dataclasses.field(default_factory=dict)


# Register a director for SubprocessInput workflow items.
# TODO: Link to deserializer behavior.
# TODO: Normalize task_builder protocol.
# TODO: Generate from class decorator.
# TODO: Need a generic TaskView class for clients with reference to managed elements.
@scalems.workflow.workflow_item_director_factory.register
def _(item: SubprocessInput, *, manager: WorkflowManager, label: str = None):
    assert isinstance(manager, WorkflowManager)

    def director(*args, **kwargs):
        if len(args) > 0:
            # TODO: Reconsider reasonable exceptions.
            raise TypeError("Unexpected positional arguments.")
        if len(kwargs) > 0:
            raise TypeError("Unexpected key word arguments: {}".format(", ".join(kwargs.keys())))
        # TODO: Checksum with scalems utility, don't just use native Python hash.
        uid = hash(item)
        if uid in manager.tasks:
            # TODO: Consider whether this is the correct behavior
            return manager.tasks[uid]
        else:
            manager.tasks[uid] = item
            return manager.tasks[uid]

    return director


@dataclasses.dataclass(frozen=True)
class SubprocessResult:
    # file: Field(Path)
    # exitcode: Field(int)
    # TODO: Can we use None instead of os.devnull to indicate non-presence of stdout/stderr?
    exitcode: int
    stdout: Path
    stderr: Path
    file: typing.Mapping[str, Path]


class SubprocessTask:
    """Describe the type of resource provided by a Subprocess command."""

    @classmethod
    def scoped_identifier(cls):
        # TODO: Consider either deriving from the `import` identifier,
        #       or defining a more sophisticated protocol for name resolution.
        return ("scalems", "subprocess", "SubprocessTask")

    @classmethod
    def identifier(cls):
        return ".".join(cls.scoped_identifier())

    @classmethod
    def input_type(cls) -> type:
        return SubprocessInput

    @classmethod
    def result_type(cls) -> type:
        return SubprocessResult


# TODO: helpers / ABCs for Serializeable and Encodable.


# TODO: Instances must have a way to map to the owning Context and a task implementation.
# It is reasonable to allow the WorkflowContext to produce an object satisfying a common
# ItemView interface, plus additional interface aspects as specified by the workflow item
# developer in e.g. SubprocessResourceType.
class Subprocess:
    @classmethod
    def resource_type(cls):
        # Note: we return an instance to better reflect the documented object
        # model and to allow for future contextual information, such as shape.
        return SubprocessTask()

    # TODO: Remove uid parameter. It should be calculated.
    def __init__(self, input: SubprocessInput, uid=None):
        self._bound_input = input
        self._result = None
        self._uid = uid
        if self._uid is None:
            self._uid = next_monotonic_integer().to_bytes(32, "big")

    def input_collection(self):
        return self._bound_input

    def result(self):
        return self._result

    def dependencies(self):
        ...

    def uid(self):
        return self._uid

    def serialize(self) -> str:
        """Encode the task as a JSON record.

        Input and Result will be serialized as references.
        The caller is responsible for serializing existing records
        for bound objects, if they exist.
        """
        record = {}
        record["uid"] = self.uid().hex()
        # "label" not yet supported.
        record["type"] = self.resource_type().scoped_identifier()
        record["input"] = dataclasses.asdict(self._bound_input)  # reference
        record["result"] = dataclasses.asdict(self._result)  # reference
        try:
            serialized = json.dumps(record, default=encode)
        except TypeError as e:
            logger.critical("Missing encoding logic for scalems data. Encoder says " + str(e))
            raise InternalError("Missing serialization support.") from e

        # raise MissingImplementationError('To do...')
        return serialized

    @classmethod
    def deserialize(cls, record: str, context=None):
        """Instantiate a Subprocess Task from a serialized record.

        In general, records should only be deserialized into a WorkflowContext
        that manages a valid work graph, but for early testing, at least,
        we have some standalone use cases.
        """

        # The record may or may not have a bound result.
        # If there is a bound result, it should be added to the workgraph first.
        # return cls()
        raise MissingImplementationError()

    # def __await__(self) -> typing.Generator[typing.Any, None, SubprocessResult]:
    #     """Implements the asyncio protocol for a coroutine object.
    #
    #     When awaited, query the current context to negotiate dispatching. Note that the
    #     built-in asyncio module acts like a LocalExecutor Context if and only if there
    #     is not an active SCALE-MS Context. SCALE-MS Contexts
    #     set the current context before awaiting.
    #     """
    #     # TODO: implementation registration.
    #     import scalems.context
    #     import scalems.local
    #     import scalems.radical
    #     context = scalems.context.get_scope()
    #     # TODO: dispatching
    #     if isinstance(context, scalems.local.LocalExecutor):
    #         from scalems.local.operations import executable as local_exec
    #         # Note that we need a more sophisticated coroutine object than what we get
    #         # directly from `async def` for command instances that can present output
    #         # in multiple contexts or be transferred from one to another.
    #         self._result = local_exec(self)
    #     elif isinstance(context, scalems.radical.RPExecutor):
    #         from scalems.radical.operations import executable as _rp_exec
    #         self._result = _rp_exec(self)
    #     else:
    #         raise MissingImplementationError(
    #         'Current context {} does not implement scalems.executable'.format(context))
    #
    #     # Allow this function to be a generator function, fulfilling the awaitable protocol.
    #     yield self
    #     # Note that the items yielded are not particularly useful, but the position of the
    #     # yield expression(s) is potentially useful for debugging or multitasking. Depending
    #     # on the implementation of the event loop, multiple yields may allow a way to avoid
    #     # deadlocks. For instance, we may choose to yield at each iteration of a loop to
    #     # provide or read PIPE-based stdin or stdout. `await` should accomplish the same thing,
    #     # but the generator protocol may improve debugging and generality.
    #     # The point of "yield" is more interesting when we use "yield" as an expression in the
    #     # yielding code, which allows values to be passed in to the coroutine at the evaluation
    #     # of the yield expression
    #     # (e.g. https://docs.python.org/3/howto/functional.html#passing-values-into-a-generator
    #     # but not that the coroutine protocol is slightly different, per https://www.python.org/dev/peps/pep-0492/)
    #     # For instance, this could be a mechanism for nesting event loops or dispatching contexts
    #     # while maintaining a heart-beat or other command-channel-like wrapper.
    #
    #     if not isinstance(self._result, SubprocessResult):
    #         raise RuntimeError('Result was not delivered!')
    #     return self._result


# Register a director for Subprocess workflow items.
# TODO: Wrap this in the decorator or metaclass used for TaskTypes.
@scalems.workflow.workflow_item_director_factory.register
def _(item: Subprocess, *, manager: scalems.workflow.WorkflowManager, label: str = None):
    if not isinstance(manager, scalems.workflow.WorkflowManager):
        raise APIError(f"No director for {repr(manager)}")

    def director(*args, **kwargs):
        if len(args) > 0:
            # TODO: Reconsider reasonable exceptions.
            raise TypeError("Unexpected positional arguments.")
        if len(kwargs) > 0:
            raise TypeError("Unexpected key word arguments: {}".format(", ".join(kwargs.keys())))

        # Note regarding registering task implementation functions:
        # scalems.subprocess is a very special kind of task. Each execution
        # environment needs to provide a specialized implementation for the
        # foreseeable future. It does not make sense for this module to provide
        # a default Python function reference for a task factory or callable.
        task_view = manager.add_item(item)

        return task_view

    return director


def executable(*args, manager: scalems.workflow.WorkflowManager = None, **kwargs):
    """Execute a command line program.

    Configure an executable to run in one (or more) subprocess(es).
    Executes when run in an execution Context, as part of a work graph.
    Process environment and execution mechanism depends on the execution environment,
    but is likely similar to (or implemented in terms of) the POSIX execvp system call.

    Shell processing of *argv* is disabled to improve determinism.
    This means that shell expansions such as environment variables, globbing (``*``),
    and other special symbols (like ``~`` for home directory) are not available.
    This allows a simpler and more robust implementation, as well as a better
    ability to uniquely identify the effects of a command line operation. If you
    think this disallows important use cases, please let us know.

    Arguments:
         manager: Workflow manager to which the work should be submitted.
         args: a tuple (or list) to be the subprocess arguments, including the executable

    *args* is required. Additional key words are optional.

    Other Parameters:
         outputs (Mapping): labeled output files, mapping command line flag to one (or
                            more) filenames.
         inputs (Mapping): labeled input files, mapping command line flag to one (or
                           more) filenames.
         environment (Mapping): environment variables to be set in the process
                                environment.
         stdin (str): source for posix style standard input file handle (default None).
         stdout (str): Capture standard out to a filesystem artifact, even if it is not
                       consumed in the workflow.
         stderr (str): Capture standard error to a filesystem artifact, even if it is
                       not consumed in the workflow.
         resources (Mapping): Name additional required resources, such as an MPI
                              environment.

    .. todo:: Support POSIX sigaction / IPC traps?

    .. todo:: Consider dataclasses.dataclass types to replace reusable/composable
              function signatures.

    Program arguments are iteratively added to the command line with standard Python
    iteration, so you should use a tuple or list even if you have only one parameter.
    I.e. If you provide a string with ``arguments="asdf"`` then it will be passed as
    ``... "a" "s" "d" "f"``. To pass a single string argument, ``arguments=("asdf")``
    or ``arguments=["asdf"]``.

    *inputs* and *outputs* should be a dictionary with string keys, where the keys
    name command line "flags" or options.

    Note that the Execution Context (e.g. RPContext, LocalContext, DockerContext)
    determines the handling of *resources*. Typical values in *resources* may include

    * procs_per_task (int): Number of processes to spawn for an instance of the *exec*.
    * threads_per_proc (int): Number of threads to allocate for each process.
    * gpus_per_task (int): Number of GPU devices to allocate for and instance of the
      *exec*.
    * launcher (str): Task launch mechanism, such as ``mpiexec``.

    Returns:
        Output collection contains *exitcode*, *stdout*, *stderr*, *file*.

    The *file* output has the same keys as the *outputs* key word argument.

    Example:
        Execute a command named ``exe`` that takes a flagged option for input
        and output file names
        (stored in a local Python variable ``my_filename`` and as the string literal
        ``'exe.out'``)
        and an ``origin`` flag
        that uses the next three arguments to define a vector.

            >>> my_filename = "somefilename"
            >>> command = scalems.executable(
            ...    ('exe', '--origin', 1.0, 2.0, 3.0),
            ...    inputs={'--infile': scalems.file(my_filename)},
            ...    outputs={'--outfile': scalems.file('exe.out')})
            >>> assert hasattr(command, 'file')
            >>> import os
            >>> assert os.path.exists(command.file['--outfile'].result())
            >>> assert hasattr(command, 'exitcode')

    TODO:
        Consider input/output files that do not appear on the command line, but which
        must figure into data flow.

    """
    if manager is None:
        manager = scalems.workflow.get_scope()

    # TODO: Figure out a reasonable way to check and catch invalid input
    #  through a dispatcher.

    # subprocess_input = context.add(Subprocess.input_type(), *args, **kwargs)
    input_type = Subprocess.resource_type().input_type()
    if not isinstance(input_type, type):
        raise InternalError(
            "Bug: {} is not coded correctly for the {}.input_type() interface.".format(
                __name__, str(type(Subprocess.resource_type()))
            )
        )

    # TODO: Add input separately. First, just add the Subprocess object.
    # Note: static type checkers may not be able to resolve that `input_type is
    # SubprocessInput` for argument checking. Provide local object to the context and
    # replace local reference with a view to the workflow item.
    # subprocess_input = _context.add_to_workflow(
    #     context, Subprocess.resource_type().input_type(), *args, **kwargs)
    bound_input = SubprocessInput(*args, **kwargs)

    director = scalems.workflow.workflow_item_director_factory(Subprocess, manager=manager)
    # Design note: at some point, dynamic workflows will require thread-safe
    # workflow editing context. We could either block on acquiring the editor
    # context, use an async context manager, or hide the possible async
    # aspect by letting the return value of the director be awaitable.

    # TODO: This would be more readable in a form like
    #  `workflow.add_item(Subprocess, bound_input)`

    try:
        task_view: scalems.workflow.ItemView = director(input=bound_input)
    except TypeError as e:
        logger.error("Invalid input in SubprocessInput: " + str(e))
        raise
    except json.JSONDecodeError as e:
        logger.critical("Malformed data: " + e.msg)
        raise InternalError("Bug: internal data is not being conditioned properly") from e
    except Exception as e:
        logger.critical("Unhandled " + repr(e))
        raise

    return task_view



"""Define the ScaleMS Subprocess command.

scalems.executable() is used to execute a program in one (or more) subprocesses.
It is an alternative to the built-in Python subprocess.Popen or asyncio.create_subprocess_exec
with extensions to better support ScaleMS execution dispatching and ensemble data flow.

The core task is represented by ``scalems.subprocess.SubprocessTask``, but also
requires the definition of ``scalems.subprocess.SubprocessInput`` and
``scalems.subprocess.SubprocessResult``.



In the first iteration, we can use dataclasses.dataclass to define input/output data structures
in terms of standard types. In a follow-up, we can use a scalems metaclass to define them
in terms of Data Descriptors that support mixed scalems.Future and native constant data types.
"""

__all__ = ["executable", "Subprocess", "SubprocessInput", "SubprocessResult"]

import logging

from . import _subprocess

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))

Subprocess = _subprocess.Subprocess
SubprocessInput = _subprocess.SubprocessInput
SubprocessResult = _subprocess.SubprocessResult

executable = _subprocess.executable



"""Tests related to RP handling of virtual environments."""
import logging
import os
import shlex
import shutil
import subprocess
import typing
from urllib.parse import ParseResult
from urllib.parse import urlparse

import packaging.version
import pytest

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@pytest.mark.experimental
def test_register_venv(rp_task_manager, rp_venv):
    """Use prepare_env() to register an existing venv."""
    import radical.pilot as rp
    import radical.utils as ru

    logger.debug(f"Client RP version is {rp.version}.")

    # We only expect one pilot
    pilot: rp.Pilot = rp_task_manager.get_pilots()[0]
    # We get a dictionary...
    # assert isinstance(pilot, rp.Pilot)
    # But it looks like it has the pilot id in it.
    pilot_uid = typing.cast(dict, pilot)["uid"]
    pmgr_uid = typing.cast(dict, pilot)["pmgr"]
    session: rp.Session = rp_task_manager.session
    pmgr: rp.PilotManager = session.get_pilot_managers(pmgr_uids=pmgr_uid)
    assert isinstance(pmgr, rp.PilotManager)
    pilot = pmgr.get_pilots(uids=pilot_uid)
    assert isinstance(pilot, rp.Pilot)
    # It looks like either the pytest fixture should deliver something other than the TaskManager,
    # or the prepare_venv part should be moved to a separate function, such as in conftest...

    access = pilot.description["access_schema"]
    if access not in ("local", "ssh"):
        pytest.skip('This test only understands "local" and "ssh" RP access schema.')

    # We use a (user-specified) venv for the Pilot agent.
    executable = os.path.join(rp_venv, "bin", "python3")

    # Create a temporary venv for this test.
    env_name = "test-env"
    env_path = "/tmp/test_env"

    scriptlet = f"test -d {env_path} && rm -rf {env_path} || true "
    scriptlet += f"; {executable} -m venv {env_path} "
    scriptlet += f"; . {env_path}/bin/activate "
    scriptlet += "; python -m pip install --upgrade pip setuptools wheel "
    # For `devel`, the reported release is not available from pypi.
    # scriptlet += f"; pip install radical.pilot=={rp.version}"
    command = ["bash", "-c", scriptlet]

    if access == "local":
        process = subprocess.run(args=command, capture_output=True, check=True, text=True, shell=False)
    else:
        # We don't have automated handling for other access methods at this time.
        assert access == "ssh"

        domain, target = str(pilot.resource).split(".")
        resource_config = ru.Config(module="radical.pilot.resource", name=domain)[target]
        ssh_target = resource_config[access]["job_manager_endpoint"]
        result: ParseResult = urlparse(ssh_target)
        assert result.scheme == "ssh"
        user = result.username
        port = result.port
        host = result.hostname

        ssh = [shutil.which("ssh")]
        if user:
            ssh.extend(["-l", user])
        if port:
            ssh.extend(["-p", str(port)])
        ssh.append(host)

        command = [shlex.quote(arg) for arg in command]
        logger.debug(f"Executing subprocess {ssh + command}")
        process = subprocess.run(
            ssh + command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=180, encoding="utf-8", shell=False
        )
        if process.returncode != 0:
            logger.error("Failed ssh stdout: " + str(process.stdout))
            logger.error("Failed ssh stderr: " + str(process.stderr))

    logger.debug(f"stdout: {process.stdout}")
    if process.stderr:
        logger.error(f"stderr: {process.stderr}")
    assert process.returncode == 0

    # Use the client-managed venv in a rp.Task with `named_env`
    # Register venv
    pilot.prepare_env(env_name=env_name, env_spec={"type": "venv", "path": env_path, "setup": []})

    # Use the registered venv.
    td = {
        "executable": "python",
        "arguments": ["-c", "import sys; print(sys.executable)"],
        "named_env": env_name,
        "pre_exec": [],
        "stage_on_error": True,
    }

    task = rp_task_manager.submit_tasks(rp.TaskDescription(td))
    task.wait(state=[rp.states.AGENT_EXECUTING] + rp.FINAL, timeout=120)
    logger.info(f"state is {task.state}.")
    logger.debug(f"stdout: {task.stdout}")
    if task.stderr:
        logger.error(f"stderr: {task.stderr}")
    assert task.exit_code == 0
    # Confirm we used the venv we think we used.
    assert env_path + "/bin/python" in task.stdout


@pytest.mark.skip(reason="Currently unused.")
def test_prepare_venv(rp_task_manager, sdist, rp_venv):
    """Bootstrap the scalems package in a RP target environment using pilot.prepare_env.

    Note that we cannot wait on the environment preparation directly, but we can define
    a task with ``named_env`` matching the *prepare_env* key to implicitly depend on
    successful creation.
    """

    import radical.pilot as rp
    import radical.saga as rs
    import radical.utils as ru

    # We only expect one pilot
    pilot: rp.Pilot = rp_task_manager.get_pilots()[0]
    # We get a dictionary...
    # assert isinstance(pilot, rp.Pilot)
    # But it looks like it has the pilot id in it.
    pilot_uid = typing.cast(dict, pilot)["uid"]
    pmgr_uid = typing.cast(dict, pilot)["pmgr"]
    session: rp.Session = rp_task_manager.session
    pmgr: rp.PilotManager = session.get_pilot_managers(pmgr_uids=pmgr_uid)
    assert isinstance(pmgr, rp.PilotManager)
    pilot = pmgr.get_pilots(uids=pilot_uid)
    assert isinstance(pilot, rp.Pilot)
    # It looks like either the pytest fixture should deliver something other than the TaskManager,
    # or the prepare_venv part should be moved to a separate function, such as in conftest...

    sdist_names = {
        "ru": ru.sdist_name,
        "rs": rs.sdist_name,
        "rp": rp.sdist_name,
        "scalems": os.path.basename(sdist),
    }
    sdist_local_paths = {"scalems": sdist, "rp": rp.sdist_path, "rs": rs.sdist_path, "ru": ru.sdist_path}
    logger.debug("Checking paths: " + ", ".join(sdist_local_paths.values()))
    for path in sdist_local_paths.values():
        assert os.path.exists(path)

    sandbox_path = urlparse(pilot.pilot_sandbox).path

    sdist_session_paths = {name: os.path.join(sandbox_path, sdist_names[name]) for name in sdist_names.keys()}

    logger.debug("Staging " + ", ".join(sdist_session_paths.values()))

    input_staging = []
    for name in sdist_names.keys():
        input_staging.append({"source": sdist_local_paths[name], "target": sdist_names[name], "action": rp.TRANSFER})
    logger.debug(str(input_staging))
    pilot.stage_in(input_staging)

    tmgr = rp_task_manager

    packages = ["pip", "setuptools", "wheel"]
    packages.extend(sdist_session_paths.values())

    # We test in multiple environments, so we have to check what Python
    # interpreter is installed in the current target resource.
    # Note that, at this time, we use a single (user-specified) venv for the
    # Pilot agent and for the tasks.
    executable = os.path.join(rp_venv, "bin", "python3")
    scriptlet = 'import sys; print(f"{sys.version_info.major}.{sys.version_info.minor}")'
    access = pilot.description["access_schema"]
    if access == "local":
        command = [executable, "-c", scriptlet]
        process = subprocess.run(args=command, capture_output=True, check=True, text=True)
    else:
        # We don't have automated handling for other access methods at this time.
        assert access == "ssh"

        domain, target = str(pilot.resource).split(".")
        resource_config = ru.Config(module="radical.pilot.resource", name=domain)[target]
        ssh_target = resource_config[access]["job_manager_endpoint"]
        result: ParseResult = urlparse(ssh_target)
        assert result.scheme == "ssh"
        user = result.username
        port = result.port
        host = result.hostname

        ssh = [shutil.which("ssh")]
        if user:
            ssh.extend(["-l", user])
        if port:
            ssh.extend(["-p", str(port)])
        ssh.append(host)

        command = [executable, "-c", shlex.quote(scriptlet)]

        logger.debug(f"Executing subprocess {ssh + command}")
        process = subprocess.run(
            ssh + command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=5, encoding="utf-8"
        )
        if process.returncode != 0:
            logger.error("Failed ssh stdout: " + str(process.stdout))
            logger.error("Failed ssh stderr: " + str(process.stderr))

    assert process.returncode == 0
    python_version = process.stdout.rstrip()
    logger.debug(f"Requesting Python version {python_version}.")
    pilot.prepare_env(
        env_name="scalems_env", env_spec={"type": "virtualenv", "version": python_version, "setup": packages}
    )

    rp_check_desc = rp.TaskDescription(
        {
            "executable": "python3",
            "arguments": ["-c", "import radical.pilot as rp; print(rp.version_detail)"],
            "named_env": "scalems_env",
        }
    )
    scalems_check_desc = rp.TaskDescription(
        {
            "executable": "python3",
            "arguments": ["-c", "import scalems; print(scalems.__file__)"],
            "named_env": "scalems_env",
        }
    )

    rp_check_task, scalems_check_task = tmgr.submit_tasks([rp_check_desc, scalems_check_desc])
    tmgr.wait_tasks()

    rp_version = rp_check_task.stdout.rstrip()
    rp_version = packaging.version.parse(rp_version)
    if rp_check_task.stdout:
        logger.debug(f"RP version details: {rp_version}")
    if rp_check_task.stderr:
        logger.debug(f"Task stderr: {rp_check_task.stderr}")
    assert rp_version == packaging.version.parse(rp.version)
    assert rp_check_task.exit_code == 0

    if scalems_check_task.stdout:
        logger.debug(f"scalems package module: {scalems_check_task.stdout}")
    if scalems_check_task.stderr:
        logger.debug(f"Task stderr: {scalems_check_task.stderr}")
    assert scalems_check_task.exit_code == 0



"""Test the scalems package data model helpers and metaprogramming utilities.
"""
import json
import logging
import os
import pathlib
import uuid

import pytest

import scalems.workflow
from scalems.exceptions import ProtocolError
from scalems.serialization import BasicSerializable
from scalems.serialization import decode
from scalems.serialization import encode
from scalems.serialization import Shape
from scalems.identifiers import TypeIdentifier

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))

record = """{
    "version"= "scalems_workflow_1",
    "types"= {
        "scalems.SubprocessInput" = {
            "schema" = {
                "spec" = "scalems.v0",
                "name" = "DataType"
            },
            "implementation" = ["scalems", "subprocess", "SubprocessInput"],
            "fields" = {
                "argv" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "String"],
                    "shape"= ["constraints.OneOrMore"]
                },
                "inputs" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "Mapping"],
                    "shape"= [1]
                },
                "outputs" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "Mapping"],
                    "shape"= [1]
                },
                "stdin" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "File"],
                    "shape"= [1]
                },
                "environment" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "Mapping"],
                    "shape"= [1]
                },
                "resources" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "Mapping"],
                    "shape"= [1]
                }
            }
        },
        "scalems.SubprocessResult" = {
            "schema" = {
                "spec" = "scalems.v0",
                "name" = "DataType"
            },
            "implementation" = ["scalems", "subprocess", "SubprocessResult"],
            "fields" = {
                "exitcode" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "Integer"],
                    "shape"= [1]
                },
                "stdout" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "File"],
                    "shape"= [1]
                },
                "stderr" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "File"],
                    "shape"= [1]
                },
                "file" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "Mapping"],
                    "shape"= [1]
                }
            }
        },
        "scalems.Subprocess" = {
            "schema" = {
                "spec" = "scalems.v0",
                "name" = "DataType"
            },
            "implementation" = ["scalems", "subprocess", "SubprocessResult"],
            "fields" = {
                "input" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "SubprocessInput"],
                    "shape"= [1]
                },
                "result" = {
                    "schema" = {
                        "spec" = "scalems.v0",
                        "name" = "DataField"
                    },
                    "type"= ["scalems", "SubprocessResult"],
                    "shape"= [1]
                }
            }
        },
    },
    "items"= [
        {
            "label"= "input_files",
            "identity"= "832df1a2-1f0b-4024-a4ab-4160717b8a8c",
            "type"= ["scalems", "Mapping"],
            "shape"= [1],
            "data"= {"-i"= ["infile"]}
        },
        {
            "label"= "output_files",
            "identity"= "d19e2734-a23c-42de-88a0-3287d7ca71ac",
            "type"= ["scalems", "Mapping"],
            "shape"= [1],
            "data"= {"-o"= ["outfile"]}
        },
        {
            "label"= "resource_spec",
            "identity"= "34dfc648-27b3-47db-b6a8-a10c9ae58f09",
            "type"= ["scalems", "Mapping"],
            "shape"= [1],
            "data"= {"ncpus"= 8, "launch_method"= ["exec"]}
        },
        {
            "label"= "subprocess_input",
            "identity"= "26c86a70-b407-471c-85ed-c1ebfa52f592",
            "type"= ["scalems", "SubprocessInput"],
            "shape"= [1],
            "data"= {
                "args"= ["myprogram", "--num_threads", "8"],
                "inputs"= "832df1a2-1f0b-4024-a4ab-4160717b8a8c",
                "outputs"= "d19e2734-a23c-42de-88a0-3287d7ca71ac",
                "stdin"= null,
                "environment" = [{}],
                "resources" = "34dfc648-27b3-47db-b6a8-a10c9ae58f09"
            },
        },
        {
            "label"= "command",
            "identity"= "199f214c-98f5-4fdd-929e-42685f8c00d2",
            "type"= ["scalems", "Subprocess"],
            "shape"= [1],
            "data"= {
                "input"= "26c86a70-b407-471c-85ed-c1ebfa52f592",
                "result"= "199f214c-98f5-4fdd-929e-42685f8c00d2"
            }
        }

    ]
}
"""


def test_shape():
    shape = Shape((1,))
    assert isinstance(shape, tuple)
    assert len(shape) == 1
    assert shape == (1,)

    shape = Shape((3, 3))
    assert isinstance(shape, tuple)
    assert len(shape) == 2
    assert shape == (3, 3)

    shape = Shape(Shape((1,)))
    assert isinstance(shape, tuple)
    assert len(shape) == 1
    assert shape == (1,)
    assert shape == Shape((1,))

    with pytest.raises(TypeError):
        # noinspection PyTypeChecker
        Shape(1)


def test_resource_type():
    scoped_name = ["scalems", "subprocess", "SubprocessTask"]
    description = scalems.workflow.Description(resource_type=TypeIdentifier(tuple(scoped_name)), shape=(1,))
    assert description.type() == TypeIdentifier(tuple(scoped_name))


def test_encoding_str():
    """Confirm that bare strings are encoded and decoded as expected.

    Note: we may choose not to support bare strings through our serialization module.
    """
    string = "asdf"
    serialized = json.dumps(string, default=encode)
    round_trip = json.loads(serialized)
    assert string == round_trip


def test_encoding_scalars():
    """Confirm that scalar data is encoded and decoded as expected.

    Note: We may choose not to support certain forms of data for the full round trip, particularly bare scalars.
    """
    scalar = 42
    serialized = json.dumps(scalar, default=encode)
    round_trip = json.loads(serialized)
    assert scalar == round_trip

    scalar = None
    serialized = json.dumps(scalar, default=encode)
    round_trip = json.loads(serialized)
    assert scalar == round_trip

    scalar = True
    serialized = json.dumps(scalar, default=encode)
    round_trip = json.loads(serialized)
    assert scalar == round_trip


def test_encoding_int():
    """Confirm that integer data is encoded and decoded as expected.

    Note: We may choose not to support certain forms of integer data for the full round trip.
    """
    series = [1, 1, 2, 3, 5]
    # length = len(series)
    # shape = (length,)

    # Test bare native int list.
    serialized = json.dumps(series, default=encode)
    round_trip = json.loads(serialized)
    assert all([a == b for a, b in zip(series, round_trip)])

    # We may expect structured data to be automatically translated to SCALE-MS representation.
    # assert round_trip.shape() == shape
    # assert round_trip.dtype() == ('scalems', 'Integer')

    array = [[1, 1], [2, 1], [8, 9]]
    # shape = (3, 2)
    serialized = json.dumps(array, default=encode)
    round_trip = json.loads(serialized)
    for rowA, rowB in zip(array, round_trip):
        assert all([a == b for a, b in zip(rowA, rowB)])
    # We expect structured data to be automatically translated to SCALE-MS representation.
    # assert round_trip.shape() == shape
    # assert round_trip.dtype() == ('scalems', 'Integer')


def test_encoding_bytes():
    length = 8
    data = (42).to_bytes(length=length, byteorder="big")
    serialized = json.dumps(data, default=encode)
    round_trip = json.loads(serialized)
    # TODO: decoder
    assert round_trip == data.hex()
    # assert data == bytes(round_trip)


def test_encoding_fileobject():
    import tempfile

    with tempfile.NamedTemporaryFile() as fh:
        filename = fh.name
        assert os.path.exists(filename)
        serialized = json.dumps(filename, default=encode)
        round_trip = json.loads(serialized)
        # TODO: decoder
        assert os.path.exists(round_trip)
        # assert round_trip.dtype() == ('scalems', 'File')
        # assert round_trip.shape() == (1,)
        assert pathlib.Path(filename) == pathlib.Path(round_trip)


def test_basic_decoding():
    # Let the basic encoder/decoder handle things that look like SCALE-MS objects.
    encoded = {
        "label": None,
        "identity": uuid.uuid4().hex,
        "type": ["test", "Spam"],
        "shape": [1],
        "data": ["spam", "eggs", "spam", "spam"],
    }
    instance = decode(encoded)
    assert type(instance) is BasicSerializable
    shape_ref = Shape((1,))
    assert instance.shape() == shape_ref
    type_ref = TypeIdentifier(("test", "Spam"))
    instance_type = instance.dtype()
    assert instance_type == type_ref

    # Test basic encoding, too.
    assert tuple(instance.encode()["data"]) == tuple(encoded["data"])
    assert instance.encode() == decode(instance.encode()).encode()
    # TODO: Check non-trivial shape.


def test_encoder_registration():
    # Test low-level registration details for object representation round trip.
    # There were some to-dos of things we should check...
    ...

    # Test framework for type creation and automatic registration.
    class SpamInstance(BasicSerializable, base_type=("test", "Spam")):
        ...

    instance = SpamInstance(
        label="my_spam",
        identity=uuid.uuid4().hex,
        dtype=["test", "Spam"],
        shape=(1,),
        data=["spam", "eggs", "spam", "spam"],
    )
    assert not type(instance) is BasicSerializable

    encoded = encode(instance)
    decoded = decode(encoded)
    assert not type(decoded) is BasicSerializable
    assert isinstance(decoded, SpamInstance)

    del instance
    del decoded
    del SpamInstance
    import gc

    gc.collect()
    with pytest.raises(ProtocolError):
        decode(encoded)

    decode.unregister(TypeIdentifier(("test", "Spam")))
    decoded = decode(encoded)
    assert type(decoded) is BasicSerializable


def test_serialization():
    ...


def test_deserialization():
    # Check for valid JSON
    serialized = json.dumps(record, default=encode)
    assert json.loads(serialized)

    ...


def test_fingerprint():
    ...


def test_object_model():
    """Check that the decorators and annotation types work right.

    Class definitions use decorators and members of special types to generate
    definitions that support the SCALE-MS object model.

    Independently from any static type checking support, this test confirms that
    the utilities actually produce the objects and relationships we expect.
    """



"""Execution dispatched through RADICAL Pilot for ScaleMS workflows.

Usage:
    python3 -m scalems.radical <RP options> my_workflow.py <script options>

"""

import sys

import scalems.invocation

# Can we attach to the rp Logger here?

if __name__ == "__main__":
    sys.exit(scalems.invocation.run(scalems.radical.workflow_manager))



"""Tools and hacks for generating unique identifiers."""
import contextvars


def next_monotonic_integer() -> int:
    """Utility for generating a monotonic sequence of integers throughout the interpreter.

    Not thread-safe. However, threads may
    * avoid race conditions by copying the contextvars context for non-root threads, or
    * reproduce the sequence of the main thread by calling this function an equal
      number of times.

    Returns:
        Next integer.

    """
    value = _monotonic_integer.get()
    _monotonic_integer.set(value + 1)
    return value


_monotonic_integer = contextvars.ContextVar("_monotonic_integer", default=0)



"""Locking protocol for scalems data stores."""

__all__ = ["LockException", "is_locked", "scoped_directory_lock"]

import contextlib
import logging
import os
import pathlib
import typing
import warnings

logger = logging.getLogger(__name__)
logger.debug("Importing {}".format(__name__))

_lock_directory_name = ".scalems_lock"
"""Name to use for lock directories (module constant)"""


class LockException(Exception):
    """The requested lock could not be obtained."""


def is_locked(path: pathlib.Path):
    """Look for evidence of a scalems lock without attempting a lock."""
    lock_names = (_lock_directory_name,)
    for name in lock_names:
        if path.joinpath(name).exists():
            return True
    return False


def _lock_directory(path=None):
    """Provide a fall-back mechanism for locking a directory.

    If we find better ways to "lock" directories in certain environments, we can decide
    how to dispatch alternative locking mechanisms. This implementation should be fairly
    robust, though, and can always serve as a fall-back.

    Locks must be explicitly removed by calling _unlock_directory(path). (Use
    `scoped_directory_lock()` for a context manager.)

    Note that, while this function should be thread-safe, the `_unlock_directory()`
    function cannot be. Care should be taken not to call `_unlock_directory()` more than
    once for a given token.

    Raises:
        LockException if the lock could not be obtained.

    """
    if path is None:
        path = os.getcwd()
    path = pathlib.Path(path).resolve()
    token_path = path.joinpath(_lock_directory_name)
    try:
        token_path.mkdir()
    except FileExistsError as e:
        # Directory lock already exists.
        raise LockException("{} already exists".format(token_path)) from e
    return token_path


def _unlock_directory(path=None):
    """Remove a lock from a directory path.

    This must be called to release a lock established with _lock_directory(path).
    """
    if path is None:
        path = os.getcwd()
    token_path = pathlib.Path(path) / _lock_directory_name
    assert token_path.exists()
    token_path.rmdir()


class _Lock:
    """Represent a lock based on a filesystem token.

    Takes ownership of a filesystem token.
    The caller is responsible for calling `.release()`
    exactly once to release the lock, removing the filesystem token.

    If the caller has not explicitly released the Lock before it is destroyed,
    a warning is issued and the __del__ method attempts to remove the object.
    """

    @property
    def name(self):
        """The name of the filesystem object provided when the lock was established."""
        return self._filesystem_object

    def __init__(self, filesystem_object=None):
        if not os.path.exists(filesystem_object):
            raise ValueError("Provided object is not a usable filesystem token.")
        self._filesystem_object = filesystem_object

    def is_active(self):
        return self._filesystem_object is not None

    def release(self):
        if self._filesystem_object is None:
            raise ValueError("Attempting to release an inactive lock.")
        _Lock._remove(self._filesystem_object)
        self._filesystem_object = None

    @staticmethod
    def _remove(path):
        """Remove a filesystem object"""
        if os.path.isfile(path):
            os.unlink(path)
        elif os.path.isdir(path):
            os.rmdir(path)
        else:
            warnings.warn("Cannot remove unknown filesystem object type: {}".format(path))

    def __del__(self):
        if self._filesystem_object is not None:
            warnings.warn(f"Lock object was not explicitly released! Token: {self._filesystem_object}")
            obj = self._filesystem_object
            self._filesystem_object = None
            if os.path.exists(obj):
                _Lock._remove(obj)


@contextlib.contextmanager
def scoped_directory_lock(path: typing.Union[str, bytes, os.PathLike] = None):
    """Create a lock directory at the provided path or in the current directory.

    Allows the caller to assert scoped control of a path through a filesystem
    token instead of inter/intra-process communication.

    Args:
        path: an existing filesystem path.

    Raises:
         LockException if the lock directory could not be created.

    Note:
        Historically, lock directories can be more effective than lock files on shared
        filesystems because of the atomicity of underlying filesystem commands. Feedback
        welcome.

    Caveats:
        * No check is currently performed on subdirectories or parents of the path
          to be locked.
        * The directory created as a lock token is assumed to remain empty.
    """
    if path is None:
        path = os.getcwd()
    token_path = _lock_directory(path)
    try:
        lock = _Lock(token_path)
    except ValueError as e:
        raise LockException("Could not create lock object.") from e
    try:
        yield lock
        # Control returns to the last line of the `try` block if the the `with`
        # block raises an exception.
    finally:
        # Note that unlink will fail if lock directory is not empty.
        # This may be a bug or a feature. Time will tell. But we should be very
        # careful about automatically removing non-empty directories on behalf
        # of users.
        if not lock.is_active():
            warnings.warn(f"Lock object {lock.name} does not exist!")
        else:
            lock.release()



"""Submodules for different simulation tools with normalized SCALE-MS friendly interface.

As we refine a normalized interface for simulation tools, we can provide
reference implementations as wrapper modules in the scalems.wrappers submodule.
Ultimately, we may be able to dispatch to the wrappers through a scalems.simulator
front end or something.

In the near term, wrappers provide a _consistent_ interface to different simulation
tools. This allows better user experience and simplifies maintenance as we
refine the software design and implement new patterns for expressing simulation
work. It allows for some high level abstraction, anticipating future functionality
for interoperability between disparate tool sets.

All wrappers should have, to the extent possible, the same methods.

Core components:
    make_input(**kwargs) -> SimulationInput:

        A simulation tool module provides a *make_input* command that accepts
        implementation-dependent named arguments and returns an object that may
        be used as simulation input within the context of the other module tools.

        Proposal: At an abstract level, this contains everything needed to define the simulation
        mathematically. The chemical system, instructions for running the simulation, but
        not how it's executed, or the resources it will take. That will be taken care of later in
        the workflow.

        GROMACS: .mpd,.top,.gro
        OpenMM: (look at ExTasy workflows)
        PLUMED-GROMACS: GROMACS + PLUMED input file
        LAMMPS: .input file

        Crucially, it provide a place for the information to be staged
        for modifying input, such that a single simulation can be
        loaded, and then variants of that simulation can be "cloned out", and modified.

        For many programs, this may simply be putting the text of the
        files into a container where they can be retrived and
        modified.

    modify_input(simulation_input: SimulationInput, **kwargs) -> SimulationInput:

        A simulation input data source produced by *make_input* or *modify_input*
        may be modified in a pipe line through (implementation dependent) key word
        arguments to a *modify_input* command.

        For most options, then a keyword argument can directly
        parallel the keyword syntax used by the simulation programs
        themselves: LAMMPS, PLUMED, GROMACS (in the .mdp) all use
        keyword options.  It's not as clear for OpenMM, but we should
        use the ExTASY implementations to guide us as to how modification is done there.

        One question is how the defintion of the chemical system
        should changed using this approach. This is not set up very
        well in the keyword format.  We could, for now, just replace
        the file containing the molecular system definition entirely
        using a modify_input.  This might be the best bet for now, but
        we will want to think about what sorts of changes might be of
        interest, if any in the future.

        The issue is that in most cases, if the chemical system is
        changing, it usually requires changes of many lines of
        topologies, and so passing it in as a dictionary will likely
        be a big pain. Likely easier to take care of it outside the
        simulation for now. Eventually, if some parallel work on
        simulation topology definitions in Open Force Field works,
        then it may be possible to do this better.

        Things like "training potential" that get attached in the
        examples should probably be attached at this point in the
        workflow, rather than at the simulate. For most workflows, it
        could ONLY be added at this stage, since potentials are
        defined in the input files.  For GMX_API driven simulations,
        the actual attachment of the potentials to the md simulation
        could happen in the simulate function, but it would be better
        if they were loaded into the data structure in the
        modify_input sage.

    simulate(simulation_input: SimulationInput, **kwargs) -> Simulation:
        Create a simulation operation instance. Accepts input from *make_input*
        or *modify_input*, along with key word arguments specified by the wrapper
        module.

        Potentially, we use this to specify which resources assoicated
        with the job?  The goal of this function is to define a task
        that is ready to dispatch for execution.

        MRS: some notes that I took during the conversation last week:
        Any of the things that affect the checkpointable identity are
        immutable after simulate.  So after simulate, define an
        equivalence class. Operate by fingerprinting commands and data
        to determine identity that should distinguish it.  Workflow
        level checkpointing.

        MRS: I'm still a little confused by the point of this.  I
        assume the importance of checkpointing is to know whether one
        should relaunch a simulation if it fails?  And if it's not
        checkpointable?  Can we clarify this a bit more.


Utility functions:
    internal_to_pdb(structure) -> scalems.FileFuture:
        For interoperability with other tools, simulator wrappers should provide
        conversion utilities for native structure and topology data. *internal_to_pdb*
        converts a native structure object to a structure file in the Protein Data Bank
        (PDB) format.

        MRS: Though actually, PDB might not be the best, since it's limited resolution.
        Should it actually convert the internal representation into a numpy array (or something
        like mdtraj or mdanalysis internal representation, which can THEN be converted into
        PDB files or any other output trajectory form desired.

    get_trajectory(source) -> Trajectory:
        Get a Future for (currently implementation-specific) Trajectory data from
        a Simulation or other sensible source. Note: this free function allows us
        to defer the question of whether trajectory sources are assumed to have a
        `trajectory` instance attribute.

    collect_coordinates(trajectories: scalems.Iterable[scalems.Iterable[Conformation]]) ->
            scalems.Iterable[Conformation]:
        Some data hierarchies or topological transformations need to be performed
        by native tools. *collect_coordinates* creates a single iterable of
        system conformations from a collection of sources of conformation data.

        MRS: Then internal_to_pdb can be used to convert into external files as needed?


Module types:
    Frame:
        Molecular system microstate data. Generally, a frame of simulation trajectory
        output, an input configuration, or a subset of simulation snapshot/checkpoint data.
        At a minimum, the member data is assumed to include atomic coordinates,
        indexed consistently with other Frames extracted from the same source.

    SimulationInput:
        Packages simulation inputs for consumption by other tools in the simulation
        package. If array-like or set-like input is provided, the SimulationInput
        has array-like or set-like behavior and implies ensemble simulation handling.
        Otherwise, the nature of the SimulationInput reference is a detail of the
        simulation tool wrapper.

    Simulation:
        Represents a simulation instance (or simulation ensemble). Named outputs
        are not yet well defined, but must include "trajectory" and something that
        is convertible to SimulationInput. The Simulation reference produced by
        a *simulate()* command has the data flow shape of the SimulationInput
        provided to it.

    Trajectory:
        A data Future representing molecular system trajectory data. For MD, this
        is presumed to be a trajectory output file or set of sequenced files.
        For a Simulation reference encompassing an ensemble or batch of simulations,
        outer dimensions of the Trajectory reference will describe the same shape
        as the source. A Trajectory has SCALE-MS sequence semantics. The wrapper
        should allow sequence-based data shaping commands to treat a Trajectory
        as a sequence of Frame data.

"""



"""Basic SCALE-MS Commands.

Core commands for building workflows.

The availability and behavior of these commands is part of the SCALE-MS
specification.
"""
# We expect these commands to be identified at the top level scalems namespace
# with e.g. ('scalems', 'logical_not'). We keep them in a submodule just
# to be tidy.

__all__ = [
    "desequence",
    "extend_sequence",
    "gather",
    "logical_and",
    "logical_not",
    "map",
    "reduce",
    "resequence",
    "scatter",
    "subgraph",
    "while_loop",
]

import typing

from scalems import exceptions


def subgraph(variables=None):
    """Allow operations to be configured in a sub-context.

    The object returned functions as a Python context manager. When entering the
    context manager (the beginning of the ``with`` block), the object has an
    attribute for each of the named ``variables``. Reading from these variables
    gets a proxy for the initial value or its update from a previous loop iteration.
    At the end of the ``with`` block, any values or data flows assigned to these
    attributes become the output for an iteration.

    After leaving the ``with`` block, the variables are no longer assignable, but
    can be called as bound methods to get the current value of a variable.

    When the object is run, operations bound to the variables are ``reset`` and
    run to update the variables.

    Example::

        @scalems.function_wrapper(output={'data': float})
        def add_float(a: float, b: float) -> float:
            return a + b

        @scalems.function_wrapper(output={'data': bool})
        def less_than(lhs: float, rhs: float, output=None):
            output.data = lhs < rhs

        subgraph = scalems.subgraph(variables={'float_with_default': 1.0, 'bool_data': True})
        with subgraph:
            # Define the update for float_with_default to come from an add_float operation.
            subgraph.float_with_default = add_float(subgraph.float_with_default, 1.).output.data
            subgraph.bool_data = less_than(lhs=subgraph.float_with_default, rhs=6.).output.data
        operation_instance = subgraph()
        operation_instance.run()
        assert operation_instance.values['float_with_default'] == 2.

        loop = scalems.while_loop(function=subgraph, condition=subgraph.bool_data)
        handle = loop()
        assert handle.output.float_with_default.result() == 6

    """
    raise exceptions.MissingImplementationError()


def logical_not(value):
    """Negate boolean inputs."""
    raise exceptions.MissingImplementationError()


def logical_and(iterable):
    """Produce a boolean value resulting from the logical AND of the elements of the input.

    The API does not specify whether the result is published before all values
    of *iterable* have been inspected, but (for checkpointing, reproducibility,
    and simplicity of implementation), the operation is not marked "complete"
    until all inputs have been resolved.
    """
    raise exceptions.MissingImplementationError()


def while_loop(*, function, condition, max_iteration=10, **kwargs):
    """Generate and run a chain of operations such that condition evaluates True.

    Returns a reference that acts like a single operation in the current
    work graph, but which is a proxy to the operation at the end of a dynamically generated chain
    of operations. At run time, *condition* is evaluated for the last element in
    the current chain. If *condition* evaluates False, the chain is extended and
    the next element is executed. When *condition* evaluates True, the object
    returned by `while_loop` becomes a proxy for the last element in the chain.

    TODO: Allow external access to intermediate results / iterations?

    Arguments:
        function: a callable that produces an instance of an operation when called (with ``**kwargs``, if provided).
        condition: a call-back that returns a boolean when provided an operation instance.
        max_iteration: execute the loop no more than this many times (default 10)

    *function* (and/or *condition*) may be a "partial function" and/or stateful function object
    (such as is produced by subgraph()) in order to retain state between iterations.

    TODO: Specify requirements of *function* and *condition*. (Picklable? scalems.Function compatible?)

    TODO: Describe how *condition* could be fed asynchronously from outside of the wrapped *function*.

    TODO: Does *condition* need to evaluate directly to a native bool, or do we
        detect whether a Future is produced, and call ``result()`` if necessary?

    Warning:
        *max_iteration* is provided in part to minimize the cost of bugs in early
        versions of this software. The default value may be changed or
        removed on short notice.

    """
    raise exceptions.MissingImplementationError()


def desequence(iterable):
    """Remove sequencing from an iterable.

    Given an input of shape (N, [M, ...]), produce an iterable of resources
    (of shape (M,...)). Ordering information from the outer dimension is lost,
    even in an ensemble scope, in which parallel data flows generally retain
    their ordering.

    This allows iterative tools (e.g. `map`) to use unordered or asynchronous
    iteration on resource slices as they become available.
    """
    raise exceptions.MissingImplementationError()


def resequence(keys, collection):
    """Set the order of a collection.

    Use the ordered list of keys to define the sequence of the elements in the
    collection.

    In addition to reordering an ordered collection, this function is useful
    for applying a sequence from one part of a work flow to data that has been
    processed asynchronously.
    """
    raise exceptions.MissingImplementationError()


def gather(iterable):
    """Convert an iterable or decomposable collection to a complete collection.

    Use to synchronize/localize data. Reference with ambiguous dimensionality or
    decomposable dimensions are converted to references with concrete dimensionality
    and non-decomposable dimensions. When the output of *gather* is used as input
    to another function, the entire results of *iterable* are available in the
    operation context.

    This function should generally not be necessary when defining a work flow,
    but may be necessary to disambiguate data flow topology, such as when N
    operations should each consume an N-dimensional resource in its entirety,
    in which case, *gather* is implicitly an *all_gather*.

    Alternatives:
        For decomposable dimensions, we may want to allow a named *scope* for
        the decomposition. Then, explicit *gather* use cases would be replaced
        by an annotation to change the *scope* for a dimension, and, as part of
        a function implementation's assertion of its decomposability, a function
        could express its ensemble/decomposition behavior in terms of a particular
        target scope of an operation.
    """
    raise exceptions.MissingImplementationError()


def scatter(iterable, axis=1):
    """Explicitly decompose data.

    For input with outer dimension size N and dimensionality D,
    or for an iterable of length N and elements with dimensionality D-1,
    produce a Resource whose outer dimensions are decomposed down to dimension
    D-*axis*.

    Note: This function is not necessary if we either require fixed dimensionality
    of function inputs or minimize implicit broadcast, scatter, and gather behavior.
    Otherwise, we will need to disambiguate decomposition.
    """
    raise exceptions.MissingImplementationError()


def reduce(function, iterable):
    """Repeatedly apply a function.

    For an Iterable[T] and a function that maps (T, T) -> T, apply the function
    to map Iterable[T] -> T. Assumes *function* is associative.

    If *iterable* is ordered, commutation of operands is preserved,
    but associativity of the generated operations is not specified.
    If *iterable* is unordered, the caller is responsible for ensuring that
    *function* obeys the commutative property.

    Compare to :py:func:`functools.reduce`
    """
    raise exceptions.MissingImplementationError()


def extend_sequence(sequence_a, sequence_b):
    """Combine sequential data into a new sequence."""
    raise exceptions.MissingImplementationError()


def map(function, iterable, shape=None):
    """Generate a collection of operations by iteration.

    Apply *function* to each element of *iterable*.

    If *iterable* is ordered, the generated operation collection is ordered.

    If *iterable* is unordered, the generated operation collection is unordered.
    """
    raise exceptions.MissingImplementationError()


class Callable(typing.Protocol):
    """This protocol describes the required function signature for a SCALE-MS command."""

    def __call__(self):
        ...


class Command(typing.Protocol):
    """Protocol describing a SCALE-MS Command."""


def command(obj: Callable) -> Command:
    """Decorate a callable to create a SCALE-MS Command."""
    raise exceptions.MissingImplementationError


def function_wrapper(output: dict = None):
    # Suppress warnings in the example code.
    # noinspection PyUnresolvedReferences
    """Generate a decorator for wrapped functions with signature manipulation.

    New function accepts the same arguments, with additional arguments required by
    the API.

    The new function returns an object with an ``output`` attribute containing the
    named outputs.

    Example:

        >>> @function_wrapper(output={'spam': str, 'foo': str})
        ... def myfunc(parameter: str = None, output=None):
        ...    output.spam = parameter
        ...    output.foo = parameter + ' ' + parameter
        ...
        >>> operation1 = myfunc(parameter='spam spam')
        >>> assert operation1.spam.result() == 'spam spam'
        >>> assert operation1.foo.result() == 'spam spam spam spam'

    Arguments:
        output (dict): output names and types

    If ``output`` is provided to the wrapper, a data structure will be passed to
    the wrapped functions with the named attributes so that the function can easily
    publish multiple named results. Otherwise, the ``output`` of the generated operation
    will just capture the return value of the wrapped function.
    """
    raise exceptions.MissingImplementationError()


def poll():
    """Inspect the execution status of an operation.

    Inspects the execution graph state in the current context at the time of
    execution.

    Used in a work graph, this adds a non-deterministic aspect, but adds truly
    asynchronous adaptability.
    """
    raise exceptions.MissingImplementationError()
